/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <math.h>
#include <stdlib.h>
#include "glview.h"
#include "game.h"

#include <qprinter.h>
#include <qtimer.h>
#include <qimage.h>
#include <qpainter.h>
#include <qdatetime.h>

#include <kstandarddirs.h>
#include <kaction.h>

#define TEXTURE_SCALE 0.5

typedef GLfloat GLfloat4[4];

void setColor(float c, float alpha = 1.0) {
    GLfloat4 grey =  { .05*c,  .05*c,  .05*c, alpha };
    GLfloat4 white = {1.0 *c, 1.0 *c, 1.0 *c, alpha };
    GLfloat4 lgrey = { .7 *c,  .7 *c,  .7 *c, alpha };

    glMaterialfv(GL_FRONT, GL_AMBIENT, grey);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, white);
    glMaterialfv(GL_FRONT, GL_SPECULAR, lgrey);
}

void vertexCallback(GLdouble * p) {
    //cout << p[0] << ":" << p[1] << ":" << p[2] << endl;
    glVertex3dv(p);
    glTexCoord2f(p[1] * TEXTURE_SCALE, p[0] * TEXTURE_SCALE);
}


void shadowMatrix(GLfloat m[4][4], GLfloat n[4], GLfloat l[4])
{
    GLfloat d = n[0]*l[0] + n[1]*l[1] + n[2]*l[2];
    GLfloat c =  - n[3] + d;

    m[0][0] = -l[0]*n[0] + c; m[1][0] = -l[0]*n[1];     m[2][0] = -l[0]*n[2]    ; m[3][0] = l[0]*n[3];
    m[0][1] = -l[1]*n[0]    ; m[1][1] = -l[1]*n[1] + c; m[2][1] = -l[1]*n[2]    ; m[3][1] = l[1]*n[3];
    m[0][2] = -l[2]*n[0]    ; m[1][2] = -l[2]*n[1];     m[2][2] = -l[2]*n[2] + c; m[3][2] = l[2]*n[3];
    m[0][3] = -n[0];        ; m[1][3] = -n[1];          m[2][3] = -n[2];        ; m[3][3] = d;
}


GLView::GLView(QWidget * parent, Game *g):
    QGLWidget(parent)
{
    game = g;
    setCaption(game->gameMoves.tagPairs["White"] + " vs " + game->gameMoves.tagPairs["Black"]);
    lightPosition[0] = 0.0;
    lightPosition[1] = 15.0;
    lightPosition[2] = -15.0;
    lightPosition[3] = 1.0;
    viewAngle = QPoint(0, 110);
    viewPos = QPoint(0, -20);
    viewDistance = 10.75;
    rotateSteps = 0;

    for (unsigned int i = 0; i < sizeof(randomtable) / sizeof(int); i++)
        randomtable[i] = 360.0 * rand()/(float)RAND_MAX;

    frameTimer = new QTimer(this);
    connect(frameTimer, SIGNAL(timeout()), SLOT(frame()));
    frameTimer->start(20);
}

GLView::~GLView() {
}

void GLView::print(QPrinter * pPrinter) {
    QPainter printpainter;
    printpainter.begin(pPrinter);

    // TODO: add your printing code here

    printpainter.end();
}

/** No descriptions */

void GLView::initializeGL() {

    GLint data;
    GLboolean hasAlpha;
    glGetBooleanv(GL_RGBA_MODE, &hasAlpha);
    qDebug("has alpha: %d", hasAlpha);
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &data);
    qDebug("max texture size: %d", data);

    glClearColor(0.0, 0.0, 0.0, 0.0);	//background black
    glShadeModel(GL_SMOOTH);	//smooth shading

    glEnable(GL_CULL_FACE);

    glEnable(GL_DEPTH_TEST);
    glDepthFunc(GL_LEQUAL);

    glEnable(GL_LIGHTING);
    GLfloat white[] = {1.0, 1.0, 1.0, 1.0};
    glLightfv(GL_LIGHT0, GL_SPECULAR, white);
    glEnable(GL_LIGHT0);

    glEnable(GL_TEXTURE_2D);
    glPolygonOffset(-2.0, 1.0);

    glMaterialf(GL_FRONT, GL_SHININESS, 100.0);
    setColor(1.0);
#ifdef GL_VERSION_1_2
    glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);
#endif

    texWwood = initTexture(locate("appdata", "wwood.jpg"));
    texBwood = initTexture(locate("appdata", "bwood.jpg"));
    texWmarble = initTexture(locate("appdata", "wmarble.jpg"));
    texBmarble = initTexture(locate("appdata", "bmarble.jpg"));

    initializeObjects();
}

void GLView::paintGL() {
    if (!isVisible()) return;
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    render();
    swapBuffers();
}

void GLView::resizeGL(int w, int h) {
    glViewport(0, 0, w, h);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    aspect = (GLfloat) w / (GLfloat) h;
    gluPerspective(45.0, aspect, 0.1, 32.0);
}

//
// Mouse Events
//

void GLView::mouseMoveEvent(QMouseEvent * e) {
    if (picking >= 0) {
        pickPos = oldpickPos + e->pos() - oldPos;
        game->visiblePieces[picking].curr = screen2board(pickPos);
        game->repaintAllViews();
    }
    if (e->stateAfter() & MidButton) {
        viewPos = oldviewPos + e->pos() - oldPos;
        repaintNeeded = TRUE;
    }
    if (e->stateAfter() & RightButton) {
        viewAngle = oldviewAngle + e->pos() - oldPos;
        repaintNeeded = TRUE;
    }
}

void GLView::mousePressEvent(QMouseEvent * e) {
    if (picking < 0)
        oldPos = e->pos();
    oldviewPos = viewPos;
    oldviewAngle = viewAngle;
    if (e->button() == Qt::LeftButton) {
        setMouseTracking(leftClick(e->pos()));
    }
}


void GLView::wheelEvent(QWheelEvent * e) {
    viewDistance -= e->delta() * 0.005;
    repaintNeeded = TRUE;
}

void GLView::renderPieces(float c, int shadow) {

    QValueList<ViewPiece>::iterator i;
    int ii;
//    glDisable(GL_BLEND);
    if (shadow) {
        glColor4f(0.0, 0.0, 0.0, c);
    } else
        setColor(c);

    for (int f=0; f<2; f++) {
        for (i = game->visiblePieces.begin(), ii=0; i != game->visiblePieces.end(); ++i, ++ii)
            if ((*i).type && (*i).fading != 0 == f) {
                glPushMatrix();
                glTranslatef((*i).curr.x, 0.0, 7 - (*i).curr.y);
                if (f)
                    if (shadow)
                        glColor4f(0.0, 0.0, 0.0, c*(1.0-(*i).alpha));
                    else
                        setColor(c, 1.0-(*i).alpha);

                if ((*i).type > 0)
                    glBindTexture(GL_TEXTURE_2D, texWwood);
                else
                    glBindTexture(GL_TEXTURE_2D, texBwood);

                glMatrixMode(GL_TEXTURE);
                glLoadIdentity();
//                glRotatef(randomtable[ abs((*i).type)*8 + (*i).idx - 8], 0.0, 0.0, 1.0);
//                glTranslatef(randomtable[ abs((*i).type)*8 + (*i).idx - 7], 0.0, 0.0);
                glMatrixMode(GL_MODELVIEW);
                glLoadName(ii);

                glCallList(dlPiece[abs((*i).type)]);
                glPopMatrix();
            }
        glEnable(GL_BLEND);
        glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    }
}

//convert board to screen coordinates
//simply draws a vertex with the board coordinates to the feedback buffer

QPoint GLView::board2screen(CoorFloat q) {
    QPoint p2;
    glFeedbackBuffer(3, GL_2D, (GLfloat *)edges);
    glRenderMode(GL_FEEDBACK);
    rotate();
    glBegin(GL_POINTS);
    glVertex3f(q.x, 0, 7-q.y);
    glEnd();
    glRenderMode(GL_RENDER);
    p2.setX((int)edges[0][1]);
    p2.setY(height()-(int)edges[0][2]);
    return p2;
}

//convert screen to board coordinates.
//basically it uses the newton method to find a solution to
//board2screen(qx, qy) - screen_coordinate = 0, if screen_coordiante is given.
//not very efficient, but i yet have to find a better method

CoorFloat GLView::screen2board(QPoint p1) {
//    qDebug("%5d %5d", p1x, p1y);
    //very first approximation
    static CoorFloat lastq(3.5, 3.5);
    
    //fist qpproximation: use the last value, since this function is
    //probably called in a mouse movement and the mouse pointer tends
    //to move only a few pixels.
    CoorFloat q = lastq;

    GLfloat d = 1.0;
    for (int j=0; j<10; j++) {
        //calculate screen coordiantes of approximation by
        //rendering to the feedback buffer
        glFeedbackBuffer(12, GL_2D, (GLfloat *)edges);
        glRenderMode(GL_FEEDBACK);
        rotate();
        glBegin(GL_POINTS);
        glVertex3f(q.x, 0, 7-q.y);
        glVertex3f(q.x+d, 0, 7-q.y);
        glVertex3f(q.x, 0, 7-q.y-d);
        glEnd();
        glRenderMode(GL_RENDER);

        //find differnce of rendered coordinates and given coordiantes
        GLfloat p2x = p1.x() - edges[0][1];
        GLfloat p2y = height()-p1.y() - edges[0][2];
        if (fabs(p2x) <= 1 && fabs(p2y) <= 1) break;

        //calculate derivations u and v in x- and z direction
        GLfloat ux = (edges[1][1] - edges[0][1])/d;
        GLfloat uy = (edges[1][2] - edges[0][2])/d;
        GLfloat vx = (edges[2][1] - edges[0][1])/d;
        GLfloat vy = (edges[2][2] - edges[0][2])/d;

        //calculate next approximation by finding the solution dq to
        //the linear combination
        //p2 = dqx * u + dqy * v
        //since u and v a vectors, we have to
        //solve two equations with two variables.
        GLfloat det = ux*vy - uy*vx;
	q = q + CoorFloat((p2x*vy - p2y*vx), (ux*p2y - uy*p2x)) * (1.0/det);
//        q.x += (p2x*vy - p2y*vx) / det;
//        q.y += (ux*p2y - uy*p2x) / det;
        q.x = 0.0 >? q.x <? 7.0;
        q.y = 0.0 >? q.y <? 7.0;
        lastq = q;
        d *= 0.5;
    }


    return q;
}

void GLView::rotate() {
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef(viewPos.x() * 0.01, -viewPos.y() * 0.01, -viewDistance);
    glRotatef(viewAngle.y() * 0.5, 1.0, 0.0, 0.0);
    glRotatef(viewAngle.x() * 0.5, 0.0, 1.0, 0.0);
    glTranslatef(-3.5, 0.0, -3.5);	//move middle of board to origin
}


void GLView::render() {
    QTime t = QTime::currentTime();
    //TODO
    //do correct lighting for reflections. real and refelcted pieces are both lit by
    //real and refelcted lightsource, but not neccessarily equal
    rotate();
    if (reflections) {
        // disable color/depth updates, o no changes do happen on the screen
        glDisable(GL_DEPTH_TEST);
        glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);

        /* Draw 1 into the stencil buffer. */
        glEnable(GL_STENCIL_TEST);
        glStencilFunc(GL_ALWAYS, 1, 0xffffffff);
        glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);

        /* Now render the parts of the board, which are to reflect the pieces later */
        glCallList(dlBoard);

        //enable color/depth updates
        glEnable(GL_DEPTH_TEST);
        glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);

        /* Now, only render where stencil is set to 1. */
        glStencilFunc(GL_EQUAL, 1, 0xffffffff);  /* draw if ==1 */
        glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);

        //reflect pieces on the xz plane. orientation of polygons has changed
        glPushMatrix();
        glScalef(1.0, -1.0, 1.0);
        glCullFace(GL_FRONT);
        glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
        glDisable(GL_BLEND);
        renderPieces(0.4, 0);

        glPopMatrix();
        glDisable(GL_STENCIL_TEST);
        glCullFace(GL_BACK);
    }

    glLightfv(GL_LIGHT0, GL_POSITION, lightPosition);
    glInitNames();
    glPushName(0xffffffff);

    //draw board
    if (shadows) {
       /* Draw the floor with stencil value 3.  This helps us only
          draw the shadow once per floor pixel (and only on the
    qDebug("render1");
          floor pixels). */
        glEnable(GL_STENCIL_TEST);
        glStencilFunc(GL_ALWAYS, 3, 0xffffffff);
        glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
    }
    glEnable(GL_BLEND);
    glBlendFunc(GL_ONE, GL_ONE);
//    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    setColor(1.0);
    glCallList(dlBoard);
    glDisable(GL_BLEND);
    glCallList(dlBoardFrame);

    GLfloat shadowMat[4][4];
//    for (int i=0; i<4; i++)
//        qDebug("%f", groundplane[i]);
    shadowMatrix(shadowMat, groundplane, lightPosition);

    glEnable(GL_BLEND);
    glBlendFunc (GL_ONE, GL_ONE);
    glCallList(dlCoor);
    glDisable(GL_BLEND);
    renderPieces(1.0, 0);

    if (shadows) {
        /* Now, only render where stencil is set above 2 (ie, 3 where
          the top floor is).  Update stencil with 2 where the shadow
          gets drawn so we don't redraw (and accidently reblend) the
          shadow). */
        glStencilFunc(GL_LESS, 2, 0xffffffff);  /* draw if ==1 */
        glStencilOp(GL_REPLACE, GL_REPLACE, GL_REPLACE);
//        glDisable(GL_STENCIL_TEST);

      /* To eliminate depth buffer artifacts, we use polygon offset
        to raise the depth of the projected shadow slightly so
        that it does not depth buffer alias with the floor. */
//         glEnable(GL_POLYGON_OFFSET_EXT);
          glEnable(GL_POLYGON_OFFSET_FILL);

      /* Render 50% black shadow color on top of whatever the
         floor appareance is. */
      glEnable(GL_BLEND);
      glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//    glBlendFunc (GL_ONE, GL_ONE);
      glDisable(GL_LIGHTING);  /* Force the 50% black. */
//      setColor(0.0, 0.5);
      glPushMatrix();
       /* Project the shadow. */
      glMultMatrixf((GLfloat *) shadowMat);
      renderPieces(0.5, 1);
      glPopMatrix();

      glDisable(GL_BLEND);
      glEnable(GL_LIGHTING);
//         glDisable(GL_POLYGON_OFFSET_EXT);
          glDisable(GL_POLYGON_OFFSET_FILL);
        glDisable(GL_STENCIL_TEST);
    }

//    qDebug("rendertime: %4d",  t.msecsTo(QTime::currentTime()));
}

GLuint GLView::initTexture(QString filename) {

    GLuint texName;

    QImage file(filename);

    QImage buf = convertToGLFormat(file);	// flipped 32bit RGBA

    glGenTextures(1, &texName);
    glBindTexture(GL_TEXTURE_2D, texName);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGB, buf.width(), buf.height(), GL_RGBA, GL_UNSIGNED_BYTE, buf.bits());
    return texName;
}

void GLView::boardEdge(GLfloat x0, GLfloat y0, GLfloat x1, GLfloat y1, GLfloat x2, GLfloat y2, GLfloat x3, GLfloat y3) {
    glNormal3f((y0 - y1) / 9.0, 0, (x1 - x0) / 9.0);
    glTexCoord2f(x0 * 0.1, 0.2 *0.1);
    glVertex3f(x0, -0.2, y0);
    glTexCoord2f(x1 *0.1, 0.2 *0.1);
    glVertex3f(x1, -0.2, y1);
    glTexCoord2f(x1 *0.1, 0 *0.1);
    glVertex3f(x1, 0, y1);
    glTexCoord2f(x0 *0.1, 0 *0.1);
    glVertex3f(x0, 0, y0);

    glNormal3f(0, 1, 0);
    glTexCoord2f(x0 *0.1, y0 *0.1);
    glVertex3f(x0, 0, y0);
    glTexCoord2f(x1 *0.1, y1 *0.1);
    glVertex3f(x1, 0, y1);
    glTexCoord2f(x2 *0.1, y2 *0.1);
    glVertex3f(x2, 0, y2);
    glTexCoord2f(x3 *0.1, y3 *0.1);
    glVertex3f(x3, 0, y3);
}

void GLView::initializeObjects() {
    //
    // Square in the middle of the board, never gets drawed
    // only for transforming screen coordinates to board coordiantes
    //
    dl4edges = glGenLists(1);
    glNewList(dl4edges, GL_COMPILE);
    glBegin(GL_POINTS);
    GLfloat v0[3] = {3, 0, 3};
    GLfloat v1[3] = {3, 0, 4};
    GLfloat v2[3] = {4, 0, 3};
    GLfloat v3[3] = {4, 0, 4};

    glVertex3fv(v0);
    glVertex3fv(v1);
    glVertex3fv(v2);
    glVertex3fv(v3);

    groundplane[0] = 0;
    groundplane[1] = 1;
    groundplane[2] = 0;
    groundplane[3] = 0;
    glEnd();
    glEndList();
    //
    // Coordinates (text)
    //
    glGenTextures(16, texCoor);
    QPixmap px(128, 256);
    QPainter p;
    QFont fnt("Times New Roman", 180);

    for (int i = 0; i < 8; i++) {
        px.fill(Qt::black);
        p.begin(&px);
        p.setFont(fnt);
        p.setPen(Qt::yellow);
        p.drawText(0, 190, QChar(char (i + '1')));
        p.end();

        QImage buf = convertToGLFormat(px.convertToImage());
        glBindTexture(GL_TEXTURE_2D, texCoor[i]);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, buf.width(), buf.height(), GL_RGBA, GL_UNSIGNED_BYTE, buf.bits());


        px.fill(Qt::black);
        p.begin(&px);
        p.setFont(fnt);
        p.setPen(Qt::yellow);
        p.drawText(0, 180, QChar(char (i + 'a')));
        p.end();

        buf = convertToGLFormat(px.convertToImage());
        glBindTexture(GL_TEXTURE_2D, texCoor[i+8]);
        glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
        gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, buf.width(), buf.height(), GL_RGBA, GL_UNSIGNED_BYTE, buf.bits());
    }

    dlCoor = glGenLists(1);
    glNewList(dlCoor, GL_COMPILE);
    setColor(1.0);
    for (int i = 0; i < 8; i++) {

        glBegin(GL_QUADS);

        glBindTexture(GL_TEXTURE_2D, texCoor[i]);
        glNormal3f(0.0, 1.0, 0.0);
        glTexCoord2f(0, 1);
        glVertex3f(-0.75, 0.001, 6.75 - i);
        glTexCoord2f(0, 0);
        glVertex3f(-0.75, 0.001, 7.25 - i);
        glTexCoord2f(1, 0);
        glVertex3f(-0.5, 0.001, 7.25 - i);
        glTexCoord2f(1, 1);
        glVertex3f(-0.5, 0.001, 6.75 - i);

        glEnd();


        glBegin(GL_QUADS);

        glBindTexture(GL_TEXTURE_2D, texCoor[i+8]);
        glNormal3f(0.0, 1.0, 0.0);
        glTexCoord2f(0, 1);
        glVertex3f(-0.125 + i, 0.001, 7.5);
        glTexCoord2f(0, 0);
        glVertex3f(-0.125 + i, 0.001, 8.0);
        glTexCoord2f(1, 0);
        glVertex3f(0.125 + i, 0.001, 8.0);
        glTexCoord2f(1, 1);
        glVertex3f(0.125 + i, 0.001, 7.5);

        glEnd();

    }
    glEndList();

    //
    //Board
    //

    dlBoardFrame = glGenLists(1);

    glNewList(dlBoardFrame, GL_COMPILE);
    glBindTexture(GL_TEXTURE_2D, texBwood);
    glBegin(GL_QUADS);
    //4 borders
    setColor(1.0);
    boardEdge(-1, 8, 8, 8, 7.5, 7.5, -.5, 7.5);
    boardEdge(8, 8, 8, -1, 7.5, -0.5, 7.5, 7.5);
    boardEdge(8, -1, -1, -1, -.5, -.5, 7.5, -.5);
    boardEdge(-1, -1, -1, 8, -.5, 7.5, -.5, -.5);
    glEnd();
    //squares
    glEndList();

    dlBoard = glGenLists(1);
    glNewList(dlBoard, GL_COMPILE);

    for (int x = 0; x < 8; x++)
        for (int y = 0; y < 8; y++) {
            if ((x + y) % 2)
                glBindTexture(GL_TEXTURE_2D, texBmarble);
            else
                glBindTexture(GL_TEXTURE_2D, texWmarble);

            glLoadName(x + (7 - y) * 8 + 0x100);
            glBegin(GL_QUADS);
            glNormal3f(0, 1, 0);
            GLfloat phi = rand()/(float)RAND_MAX * M_PI * 2;
            GLfloat tx = cos(phi) * 0.25;
            GLfloat ty = sin(phi) * 0.25;
            GLfloat xx = rand()/(float)RAND_MAX*0.5+0.25;     //0.25..0.75
            GLfloat yy = rand()/(float)RAND_MAX*0.5+0.25;     //0.25..0.75
            glTexCoord2f(tx + xx,   ty + yy);
            glVertex3f  ( x + 0.5, 0, y + 0.5);
            glTexCoord2f(ty + xx,  -tx + yy);
            glVertex3f  ( x + 0.5, 0, y - 0.5);
            glTexCoord2f(-tx+ xx , -ty + yy);
            glVertex3f  ( x - 0.5, 0, y - 0.5);
            glTexCoord2f(-ty+ xx,   tx + yy);
            glVertex3f  ( x - 0.5, 0, y + 0.5);
            glEnd();
        }
    glEndList();

    for (int i = 1; i < 7; i++)
        dlPiece[i] = glGenLists(1);

    GLfloat kheadx[] = { -0.05, 0.05, 0.05, 0.15, 0.15, 0.05, 0.05, -0.05, -0.05, -0.15,
                         -0.15, -0.05
                       };
    GLfloat kheady[] = { 2.050, 2.050, 2.150, 2.150, 2.250, 2.250, 2.350, 2.350, 2.250,
                         2.250, 2.150, 2.150
                       };
    //-0.15,-0.15,-0.05,-0.05, 0.05, 0.05, 0.15,  0.15,  0.05,  0.05, -0.05, -0.05};
    int kheadc = sizeof(kheadx) / sizeof(GLfloat);

    GLfloat knheadx[] = { 0.2, 0.286, 0.0429, 0.229, 0.257, 0.357, 0.357, 0.271, 0.0714,
                          -0.0857, -0.0857, -0.114, -0.143, -0.271, -0.314, -0.329, -0.314,
                          -0.286,
                          -0.257, -0.214, -0.129
                        };
    //{41,47,30,43,45,52,52,46,32,21,21,19,17, 8, 5, 4, 5, 7, 9,12,18};
    GLfloat knheady[] = { 0.425, 0.468, 0.911, 0.939, 0.911, 0.996, 1.111, 1.139, 1.282,
                          1.268, 1.354, 1.382, 1.382, 1.239, 1.096, 0.911, 0.611, 0.496,
                          0.468,
                          0.468,
                          0.425
                        };
    //72,69,38,36,38,32,24,22,12,13, 7, 5, 5,15,25,38,59,67,69,69,72};
    int knheadc = sizeof(knheadx) / sizeof(GLfloat);

    GLfloat knightx[] = { 0.0, 1.3 / 4, 1.3 / 4, 1.2 / 4, 1.3 / 4, 1.2 / 4, 0.95 / 4,
                          0.9 / 4, 0.0
                        };
    GLfloat knighty[] = { 0.0, 0.0 / 4, 0.3 / 4, 0.4 / 4, 0.7 / 4, 1.0 / 4, 1.4 / 4,
                          1.7 / 4, 0.425
                        };
    int knightc = sizeof(knightx) / sizeof(GLfloat);

    GLfloat rookx[] = { 0.0 / 4, 1.3 / 4, 1.3 / 4, 1.2 / 4, 1.3 / 4, 1.2 / 4, 0.95 / 4,
                        0.9 / 4, 0.65 / 4, 0.15, 0.25, 0.275, 0.275, 0.25, 0.25, 0.275,
                        0.3, 0.25,
                        0.225, 0
                      };
    GLfloat rooky[] = { 0.0 / 4, 0.0 / 4, 0.3 / 4, 0.4 / 4, 0.7 / 4, 1.0 / 4, 1.4 / 4,
                        1.7 / 4, 2.0 / 4, 0.825, 0.925, 0.925, 0.95, 0.95, 0.975, 0.975,
                        1.3, 1.3,
                        1.175, 1.175
                      };
    int rookc = sizeof(rookx) / sizeof(GLfloat);

    GLfloat queenx[] = { 0.0 / 4, 1.6 / 4, 1.6 / 4, 1.5 / 4, 1.6 / 4, 1.6 / 4, 1.2 / 4,
                         1.15 / 4, 0.8 / 4, 0.6 / 4, 0.5 / 4, 1.0 / 4, 1.2 / 4, 0.9 / 4,
                         0.85 / 4,
                         0.7 / 4, 0.7 / 4, 0.8 / 4, 0.7 / 4, 0.8 / 4, 1.25 / 4, 1.1 / 4,
                         0.8 / 4,
                         0.35 / 4, 0.4 / 4, 0.4 / 4, 0.3 / 4, 0.0
                       };
    GLfloat queeny[] = { 0.0 / 4, 0.0 / 4, 0.4 / 4, 0.5 / 4, 0.8 / 4, 1.0 / 4, 1.7 / 4,
                         2.2 / 4, 2.5 / 4, 3.5 / 4, 5.0 / 4, 5.1 / 4, 5.3 / 4, 5.5 / 4,
                         5.7 / 4,
                         5.8 / 4, 6.0 / 4, 6.2 / 4, 6.4 / 4, 6.9 / 4, 7.7 / 4, 8.0 / 4,
                         7.9 / 4,
                         8.3 / 4, 8.6 / 4, 8.7 / 4, 8.8 / 4, 2.2
                       };
    int queenc = sizeof(queenx) / sizeof(GLfloat);

    GLfloat kingx[] = { 0.0 / 4, 1.6 / 4, 1.6 / 4, 1.5 / 4, 1.6 / 4, 1.6 / 4, 1.2 / 4,
                        1.15 / 4, 0.8 / 4, 0.6 / 4, 0.5 / 4, 1.0 / 4, 1.2 / 4, 0.9 / 4,
                        0.85 / 4,
                        0.7 / 4, 0.7 / 4, 0.8 / 4, 0.7 / 4, 0.8 / 4, 1.25 / 4, 1.1 / 4,
                        0.8 / 4,
                        0.35 / 4,
                        0.0
                      };
    GLfloat kingy[] = { 0.0 / 4, 0.0 / 4, 0.4 / 4, 0.5 / 4, 0.8 / 4, 1.0 / 4, 1.7 / 4,
                        2.2 / 4, 2.5 / 4, 3.5 / 4, 5.0 / 4, 5.1 / 4, 5.3 / 4, 5.5 / 4,
                        5.7 / 4,
                        5.8 / 4, 6.0 / 4, 6.2 / 4, 6.4 / 4, 6.9 / 4, 7.7 / 4, 8.0 / 4,
                        8.2 / 4,
                        8.3 / 4,
                        2.125
                      };
    int kingc = sizeof(kingx) / sizeof(GLfloat);

    GLfloat bishopx[] = { 0.0 / 4, 1.3 / 4, 1.3 / 4, 1.2 / 4, 1.3 / 4, 1.2 / 4, 0.95 / 4,
                          0.9 / 4, 0.65 / 4, 0.45 / 4, 1.0 / 4, 1.1 / 4, 1.0 / 4, 0.6 / 4,
                          0.9 / 4,
                          0.95 / 4, 0.9 / 4, 0.65 / 4, 0.25 / 4, 0.4 / 4, 0.4 / 4, 0.3 / 4,
                          0.0
                        };
    GLfloat bishopy[] = { 0.0 / 4, 0.0 / 4, 0.3 / 4, 0.4 / 4, 0.7 / 4, 1.0 / 4, 1.4 / 4,
                          1.7 / 4, 2.0 / 4, 3.0 / 4, 3.1 / 4, 3.3 / 4, 3.5 / 4, 3.7 / 4,
                          4.5 / 4,
                          4.75 / 4, 5.0 / 4, 5.5 / 4, 6.2 / 4, 6.3 / 4, 6.4 / 4, 6.5 / 4,
                          1.625
                        };
    int bishopc = sizeof(bishopx) / sizeof(GLfloat);

    GLfloat pawnx[] = { 0.000, 0.303, 0.289, 0.261, 0.267, 0.206, 0.220, 0.138, 0.088, 0.069, 0.138, 0.179, 0.110, 0.138, 0.165, 0.138, 0.069, 0.000 };
    GLfloat pawny[] = { 0.000, 0.000, 0.069, 0.110, 0.206, 0.275, 0.357, 0.413, 0.495, 0.619, 0.619, 0.688, 0.756, 0.825, 0.921, 1.018, 1.073, 1.073 };

    int pawnc = sizeof(pawnx) / sizeof(GLfloat);
    
    glNewList(dlPiece[KNIGHT], GL_COMPILE);
    revolve(knightx, knighty, knightc);
    extrude(knheadx, knheady, knheadc, 0.1);
    glEndList();

    glNewList(dlPiece[PAWN], GL_COMPILE);
    revolve(pawnx, pawny, pawnc);
    glEndList();

    glNewList(dlPiece[KING], GL_COMPILE);
    revolve(kingx, kingy, kingc);
    extrude(kheadx, kheady, kheadc, 0.05);
    glEndList();

    glNewList(dlPiece[BISHOP], GL_COMPILE);
    revolve(bishopx, bishopy, bishopc);
    glEndList();

    glNewList(dlPiece[QUEEN], GL_COMPILE);
    revolve(queenx, queeny, queenc);
    glEndList();

    glNewList(dlPiece[ROOK], GL_COMPILE);
    revolve(rookx, rooky, rookc);
    glEndList();
}

int GLView::selection(QPoint p) {
#define BUFSIZE 512
    GLuint selectBuf[BUFSIZE];
    GLint hits, selection;
    GLint viewport[4];
    unsigned j;
    GLuint names, *ptr, zmin, z;
    glGetIntegerv(GL_VIEWPORT, viewport);

    glSelectBuffer(BUFSIZE, selectBuf);
    glRenderMode(GL_SELECT);

    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();
    /*  create 5x5 pixel picking region near cursor location      */
    gluPickMatrix((GLdouble) p.x(), (GLdouble) (viewport[3] - p.y()), 2.0, 2.0, viewport);
    gluPerspective(45.0, aspect, 0.1, 32.0);
    render();

    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glFlush();

    hits = glRenderMode(GL_RENDER);
    ptr = selectBuf;
    zmin = 0xffffffff;
    selection = -1;

    for (int i = 0; i < hits; i++) {	/*  for each hit  */
        names = ptr[0];
        z = ptr[1] + ptr[2];
        for (j = 0; j < names; j++) {	/*  for each name */
            if (z < zmin) {
                zmin = z;
                selection = ptr[j + 3];
            }

        }
        ptr += names + 3;
    }

    if (selection >= 0) {
        QValueList<ViewPiece>::iterator i;
        for (i = game->visiblePieces.begin(); i != game->visiblePieces.end(); i++)
            if ((*i).type && selection == (*i).curr.x + (*i).curr.y * 8) {
                //		selection = i;
                break;
            }
    }

    return selection;
}

void GLView::revolve(GLfloat * xx, GLfloat * yy, int count) {

    int i;

    GLfloat dx, dy, dd, nx, ny, dxold, dxnew, dyold, dynew, ddold, ddnew;

    for (i = 1; i < count - 1; i++) {
        if (i == count - 2)
            glBegin(GL_POLYGON);
        else
            glBegin(GL_QUAD_STRIP);

        for (float phi = 0.0; phi < 2 * M_PI + 0.01; phi += M_PI / 12.0) {

            dxold = xx[i] - xx[i - 1];
            dyold = yy[i] - yy[i - 1];
            ddold = sqrt(dxold * dxold + dyold * dyold);
            dxold /= ddold;
            dyold /= ddold;

            dxnew = xx[i + 1] - xx[i];
            dynew = yy[i + 1] - yy[i];
            ddnew = sqrt(dxnew * dxnew + dynew * dynew);
            dxnew /= ddnew;
            dynew /= ddnew;

            if (dxnew * dxold + dynew * dyold < 0.5) {
                dx = dxnew;
                dy = dynew;
            } else {
                dx = dxold + dxnew;
                dy = dyold + dynew;
            }
            dd = sqrt(dx * dx + dy * dy);
            nx = dy / dd;
            ny = -dx / dd;

            if (i < count - 2) {
                glTexCoord2f(phi * (0.5/M_PI) * TEXTURE_SCALE, yy[i] * (0.5/M_PI) * TEXTURE_SCALE);
                glNormal3f(cos(phi) * nx, ny, sin(phi) * nx);
                glVertex3f(xx[i] * cos(phi), yy[i], xx[i] * sin(phi));

                dxold = xx[i + 1] - xx[i];
                dyold = yy[i + 1] - yy[i];
                ddold = sqrt(dxold * dxold + dyold * dyold);
                dxold /= ddold;
                dyold /= ddold;

                dxnew = xx[i + 2] - xx[i + 1];
                dynew = yy[i + 2] - yy[i + 1];
                ddnew = sqrt(dxnew * dxnew + dynew * dynew);
                dxnew /= ddnew;
                dynew /= ddnew;

                if (dxnew * dxold + dynew * dyold < 0.5) {
                    dx = dxold;
                    dy = dyold;
                } else {
                    dx = dxold + dxnew;
                    dy = dyold + dynew;
                }
                dd = sqrt(dx * dx + dy * dy);
                nx = dy / dd;
                ny = -dx / dd;
                glTexCoord2f(phi * (0.5/M_PI) * TEXTURE_SCALE, yy[i + 1] * (0.5/M_PI) * TEXTURE_SCALE);
                glNormal3f(cos(phi) * nx, ny, sin(phi) * nx);
                glVertex3f(xx[i + 1] * cos(phi), yy[i + 1], xx[i + 1] * sin(phi));
            } else {
                glTexCoord2f(xx[i] * cos(phi) * TEXTURE_SCALE, -xx[i] * sin(phi) * TEXTURE_SCALE);
                glNormal3f(cos(phi) * nx, ny, -sin(phi) * nx);
                glVertex3f(xx[i] * cos(phi), yy[i], -xx[i] * sin(phi));
            }

        }
        glEnd();
    }
}

void GLView::extrude(GLfloat * x, GLfloat * y, int c, GLfloat d) {
    typedef GLdouble GLdouble3[3];

    GLdouble3 *knheadVertex = new GLdouble3[c];
    int i;

    GLUtesselator *tobj = gluNewTess();
    gluTessCallback(tobj, GLU_TESS_VERTEX, (GLvoid(*)())vertexCallback);
    gluTessCallback(tobj, GLU_TESS_BEGIN, (GLvoid(*)())glBegin);
    gluTessCallback(tobj, GLU_TESS_END, (GLvoid(*)())glEnd);
    //    gluTessProperty(tobj, GLU_TESS_WINDING_RULE,  GLU_TESS_WINDING_POSITIVE);

    gluTessNormal(tobj, 0.0, 0.0, -1.0);
    glNormal3f(0.0, 0.0, -1.0);

    gluTessBeginPolygon(tobj, NULL);
    gluTessBeginContour(tobj);
    for (i = 0; i < c; i++) {
        knheadVertex[i][0] = x[i];	//(knheadx[i]-27.0)/70.0;
        knheadVertex[i][1] = y[i];	//(72.0-knheady[i])/70.0 + knighty[knightc-1];
        //      printf("%2.03f, ", y[i]+0.1);
        knheadVertex[i][2] = -d;
        gluTessVertex(tobj, knheadVertex[i], knheadVertex[i]);
    }
    //    printf("\n");
    gluTessEndContour(tobj);
    gluTessEndPolygon(tobj);

    gluTessNormal(tobj, 0.0, 0.0, 1.0);
    glNormal3f(0.0, 0.0, 1.0);

    gluTessBeginPolygon(tobj, NULL);
    gluTessBeginContour(tobj);
    for (i = 0; i < c; i++) {
        knheadVertex[i][2] = d;
        gluTessVertex(tobj, knheadVertex[i], knheadVertex[i]);
    }
    gluTessEndContour(tobj);
    gluTessEndPolygon(tobj);

    gluDeleteTess(tobj);

    GLfloat dx, dy, dd, dsum = 0;

    glBegin(GL_QUAD_STRIP);
    for (i = 0; i < c; i++) {
        if (i > 0) {
            dx = knheadVertex[i - 1][0] - knheadVertex[i][0];
            dy = knheadVertex[i - 1][1] - knheadVertex[i][1];
        } else
            dx = dy = 0;

        if (i < c - 1) {
            dx += knheadVertex[i][0] - knheadVertex[i + 1][0];
            dy += knheadVertex[i][1] - knheadVertex[i + 1][1];
        }
        dd = sqrt(dx * dx + dy * dy);
        glNormal3f(-dy / dd, dx / dd, 0.0);
        glTexCoord2f(dsum * TEXTURE_SCALE, d * TEXTURE_SCALE);
        glVertex3f(knheadVertex[i][0], knheadVertex[i][1], d);
        glTexCoord2f(dsum * TEXTURE_SCALE, -d * TEXTURE_SCALE);
        glVertex3f(knheadVertex[i][0], knheadVertex[i][1], -d);
        dsum += dd;
    }
    glEnd();
}

void GLView::frame() {
    if (repaintNeeded) {
	if (rotateSteps) {
	    viewAngle -= QPoint(19-abs(rotateSteps-18),0);
	    rotateSteps--;
	}
	paintGL();
        repaintNeeded = rotateSteps;
    }
}

void GLView::rotate180() {
    rotateSteps = 36;
    repaintNeeded = true;
}

#include "glview.moc"

