#define _FILE_OFFSET_BITS 64
#define MALLOC_CHECK_

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <climits>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <fnmatch.h>
#include <dirent.h>
#include "endgame.h"
/*
    simple approach: for a n-endgame 2 tables[64][64]...[64][64], with n indices, one for white to move
    and one for black, each element of the table has to be large enough to hold m values,
    if m is the maximum distance in moves to win. a 6-piece endgame would need 137 GByte RAM.
    Pawn endgames need 4 times the memory, because of the weaker symmetries, endgames with 2 pieces
    of the same kind need half the mem.
    
    first optimization: use the symmetries of the board and the fact, that a position,
    where the distance between both kings is less  than 2, is illegal. with the white
    king reduced to the triangle a1-d1-d4-a1 and the black to a1-h1-h8-a1, if the wk
    is on the diagonal a1-h8, we need 2 tables[462][64]... with n-1 indices, since there
    are only 462 legal positions for both kings with these constraints. the table
    can be further reduced, if there are 2 pieces of the same kind in the endgame: a simple
    sorting of the two pieces reduces a [64][64] in the table to a [2080], since there
    are only 2080 not-ordered positions of 2 pieces on the board.
    a 6 piece endgame would need 15 GByte on disk and ram.
    
    second optimization: don't store the distance to win for a position, only a bit
    if a position is winning. for chess programs, the distance-to-win table is 
    of use, only if the position is in fact played. but most accesses to a endgame database
    happen in calculation of possible variants, not in actually playing the endgame.
    in this case, knowing if a variant wins or not, is sufficient, knowing how fast
    it wins, is optional.
    a distance to win table can be derived later, after finishing the calculation. this
    reduces ram memory consumption to 1,9 GByte for a 6-piecer.
    
    third optimization: dont't hold both tables in memory. typically, one table is read
    and written with linear access (since we scan the possible positions one after another),
    the other one is read with random access. the table
    with linear access doesn't need to be in memory, since this operation can
    be performed as an file i/o. with some buffering, the speed loss is neglectable.
    and the random access isn't really 'random': one defines a page as the table containing
    all possible piece positions to a given two-kings position. for instance, in the case of
    4 pieces + 2 kings with full sym there are 462 pages of [64][64][64][64] bits.
    since a king can only move to 8 different positions, it is sufficient to hold
    only 9 pages in memory. page misses only occur when the scanning position moves
    to a new page, and this happens "only" 462 times during a scan. and since the king
    only moves to adjacent squares, most of the time we only have to load 3 new pages
    on a scanning page change. so, considering we have to read and to write a page
    on each scanning page change anyway, this increases the file i/o by 5 times, since
    we need to read and write like 4 more pages on each scan page change.
    this is further reducing the memory consumption
    to 18 MByte for a 6-piecer and 1,2 GByte for a 7-piece endgame (!). further reducing
    is possible, if one creates sub-pages, but this will increase file i/o.
    however, memory required on disk stays the same, for a 7-piecer 124 GByte for
    the bit-tables (white and black to move) and 1 T for the distance-to-win table.
    a very "easy" 8-piece endgame, with 2 times 2 pieces of the same kind and
    the constraint that bishops are on different colored squares, like KBBR vs kbbn
    could be possible to, since it only needs 4.5 GByte (compared to 72 GB if all pieces
    are of different kind), but this still requires a 64 bit architecture, 4 T diskspace
    and probably some years of calculation time.
    
*/

#define MAXPIECES 7
#define MAXPAGES 1806    //462 for full 8 fold sym, 1806 for 2 fold sym

#define BK 1
#define WK 0

#define XFLIP 1
#define YFLIP 2
#define DFLIP 4

#define ROOK 1
#define BISHOP 2
#define QUEEN 3
#define KNIGHT 4
#define PAWN 5
#define KING 6

#define WHITE 1
#define BLACK -1

#define MATE_VALUE -512

#define ITERATE(eg) for ((eg)->startpos(); !(eg)->endpos(); (eg)->nextpos())
#define DIST(piece, pos) dist[(piece)+6][(pos)+0x100]

#define CNT(x)      piecelist[6+(x)].cnt        //anzahl einer fiugren art
#define POS(x,y)    piecelist[6+(x)].pos[y]     //position der y-ten figur von typ x

using namespace std;

typedef char str[256];

typedef struct {
    int wk;
    int bk;
} KingsPosition;

typedef struct {
    int index;
    int symops;
} Page;

typedef enum {one, nine, full} Swaptype;

//globals

//arrays of all directions for sliding and hopping pieces
static const int dirtab[] = { 1, 0x11, 0x10, 0xf, -1, -0x11, -0x10, -0xf };
static const int kdirtab[] = { 0x21, 0x12, 0xe, 0x1f, -0x21, -0x12, -0xe, -0x1f };
//there are 2 basic types of endgames, with and without pawns
//they have different symmetries and differ in the number of pages
KingsPosition posTable[2][MAXPAGES];//array of all legal WK and BK positions, without symmetric pos
Page pageDir[2][0x80][0x80];        //the inverse: index into the pagetable for a WK/BK pos
unsigned pagecnt[2];                //number of pages


int dist[13][0x200];
int mirrortab[0x80][8];
bool diagonal[0x77];
bool insertpos, insertneg;

class Endgame;
Endgame *egCollection[100], **egEnd;

void sort(char *);

struct {
    bool newstart;
    bool verbose;
    bool usepaging;
    bool dtable;
    bool rtable;
} opts;
    
struct Swapfile {
    union {
        char *pageTable[MAXPAGES];      //if type == full, we have the memory addresses belonging to a page stored here
        int pageIndexTable[MAXPAGES];   //else we get a index into pageAge[] and pageIndex[], and the mem addr is
    };                                  //addr + index * pagesize
    str name;
    int pageindex[9];                   //the index into pageIndexTable. this is the inverse to pageIndexTable, so
    char *addr;                         //pageIndexTable[pageIndex[x]] == x is always true.
    unsigned pageage[9];                //access time, for the LRU algorithm
    int date;                           //the 'time' we use for pageAge
    off_t pagesize;               
    unsigned sym, cnt;                       //1 if we have no pawns and full symmetry
    int sametype[MAXPIECES];
    bool readonly;
    FILE *fp;
    Swaptype swaptype;

    void alloc(bool ro, Swaptype type) {
        readonly = ro;
        swaptype = type;

        if (readonly) {
            //if we open a swapfile for reading, it must exist
            fp = fopen(name, "r");
            if (!fp)
                fprintf(stderr, "can not open %s for reading\n", name);
        } else {
            //try to open file for reading and writing
            fp = fopen(name, "r+");
            if (!fp) {//file does not exist, create it
                fp = fopen(name, "w+");
                void *empty = calloc(1, pagesize);
                for (unsigned i=0; i<pagecnt[sym]; i++)
                    if (fwrite(empty, 1, pagesize, fp) < pagesize) {
                        fprintf(stderr, "can't write pagefile (not enough diskspace?)\n");
                        exit(1);
                    }
                ::free(empty);
            }
            if (!fp)
                fprintf(stderr, "can not open %s for writing\n", name);
        }

        switch(swaptype) {
        case one:
            addr = (char*)calloc(1, pagesize);
            pageindex[0] = -1;
            for (unsigned x=0; x<pagecnt[sym]; x++)
                pageIndexTable[x] = -1;
            break;
        case nine:
            addr = (char*)calloc(9, pagesize);
            for (int x=0; x<9; x++) {
                pageage[x] = 0;
                pageindex[x] = -1;
            }
            date = 1;
            for (unsigned x=0; x<pagecnt[sym]; x++)
                pageIndexTable[x] = -1;
            break;
        case full:
            addr = (char*)calloc(pagecnt[sym], pagesize);
            if (fp)
                fread(addr, pagecnt[sym], pagesize, fp);
/*            for (unsigned x=0; x<pagecnt[sym]; x++)
                pageTable[x] = addr + x*pagesize;*/
        }
    }

    void alloc(const char *fname, bool ro, Swaptype type, int st[MAXPIECES], int c, unsigned s) {
        strcpy(name, fname);
        cnt = c;
        sym = s;
        for (unsigned i=0; i<cnt; i++)
            sametype[i] = st[i];
            
        pagesize = 1 << (6*(cnt-2) - 3);
        
        for (unsigned i=2; i<cnt; i++)
            if (sametype[i]) {
//                if (opts.verbose) printf("Same piece type %c for piece %d to %d\n", s[i], i, i+sametype[i]);
                pagesize *= 64+sametype[i];
                pagesize /= (sametype[i]+1)*64;
            }
        alloc(ro, type);
    }
    
    void free() {
        if (fp) {
            if (!readonly) {
                switch(swaptype) {
                case one:
                    if (pageindex[0]>=0) {
                        fseeko(fp, pageindex[0]*pagesize, SEEK_SET);
                        fwrite(addr, 1, pagesize, fp);
                    }
                    break;
                case nine:
                    for (int x=0; x<9; x++) {
                        if (pageindex[x]>=0) {
                            fseeko(fp, pageindex[x]*pagesize, SEEK_SET);
                            fwrite(addr + x*pagesize, 1, pagesize, fp);
                        }
                    }
                    break;
                case full:
                    rewind(fp);
                    fwrite(addr, pagecnt[sym], pagesize, fp);
                }
            }
            fclose(fp);
        }
        ::free(addr);
    }

    void realloc(bool ro, Swaptype type) {
        free();
        alloc(ro, type);
    }    
    
#if 0
    unsigned bitadrfunc3() {
        return pos[2];
    }
    unsigned bitadrfunc4() {
        return pos[2]*64 + pos[3];
    }
    unsigned bitadrfunc5() {
        return (pos[2]*64 + pos[3])*64 + pos[4];
    }
    unsigned bitadrfunc6() {
        return ((pos[2]*64 + pos[3])*64 + pos[4])*64 + pos[5];
    }
    unsigned bitadrfunc31() {//assuming pos[3] <= pos[2]
        return pos[2]*(pos[2]+1)/2 + pos[3];
    }
    unsigned bitadrfunc41() {//assuming pos[3] <= pos[2]
        return pos[2]*(pos[2]+1)*32 + pos[3]*64 + pos[4];
    }
    unsigned bitadrfunc51() {//assuming pos[3] <= pos[2]
        return (pos[2]*(pos[2]+1)*32 + pos[3]*64 + pos[4])*64 + pos[5];
    }
#endif
/*  calculates the address of the bit corresponding to the current position.
    takes pos[2]...pos[cnt-1] as input, and mirrors the position according to
    flags. pieces of the same kind are ordered so that the piece with the higher
    index has always the lower position.

      unsigned bitadr = 1;
        for (int i=2; i<cnt; i++)
            if (sametype[i]) {
                bitadr = (64*65/2)* bitadr + (mirrortab[pos[i]][flags]+1)*mirrortab[pos[i]][flags]/2 + mirrortab[pos[i+1]][flags];
                i++;
            }
            else
                bitadr = 64* bitadr + mirrortab[pos[i]][flags];
        unsigned bitadr = mirrortab[pos[2]][flags];
        for (int i=3; i<cnt; i++)
            if (sametype[i-1])
                bitadr = (mirrortab[pos[i-1]][flags]+1)*bitadr/2 + mirrortab[pos[i]][flags];
            else
                bitadr = 64* bitadr + mirrortab[pos[i]][flags];
*/    
    unsigned getbitadr(unsigned flags, unsigned pos[]) {
        static bool swap[MAXPIECES] = {false,false,false,false,false,false,false};
        for (unsigned i=2; i<cnt-1; i++)
            if (sametype[i])
                if (swap[i] = mirrortab[pos[i+1]][flags] > mirrortab[pos[i]][flags]) {
                    int temp = pos[i];
                    pos[i] = pos[i+1];
                    pos[i+1] = temp;
                }

        unsigned bitadr = 0;
        for (unsigned i=2; i<cnt; i++)
            if (sametype[i]) {
                bitadr = (64*65/2)* bitadr + (mirrortab[pos[i]][flags]+1)*mirrortab[pos[i]][flags]/2 + mirrortab[pos[i+1]][flags];
                i++;
            }
            else
                bitadr = 64* bitadr + mirrortab[pos[i]][flags];

        for (unsigned i=2; i<cnt; i++)
            if (swap[i]) {
                int temp = pos[i];
                pos[i] = pos[i+1];
                pos[i+1] = temp;
                swap[i] = false;
            }

        return bitadr;
    }

    off_t linearaddr(unsigned pos[]) {
        return ((off_t)getbitadr(pageDir[sym][pos[BK]][pos[WK]].symops, pos)
            + (off_t)pageDir[sym][pos[BK]][pos[WK]].index*pagesize*8) * 2;
    }
    
    char *byteaddr(unsigned pos[]) {
        if (pageDir[sym][pos[BK]][pos[WK]].index < 0)
            fprintf(stderr, "Illegal WK/BK\n");
        switch (swaptype) {
        default:
        case full:
//            return pageTable[pageDir[sym][pos[BK]][pos[WK]].index];
            return addr+pagesize*pageDir[sym][pos[BK]][pos[WK]].index;
        case one:
            if (pageIndexTable[pageDir[sym][pos[BK]][pos[WK]].index]>=0)
                return addr;
            if (pageindex[0] >= 0) {
                if (!readonly) {
                    fseeko(fp, pageindex[0]*pagesize, SEEK_SET);
                    fwrite(addr, 1, pagesize, fp);
                }
                pageIndexTable[pageindex[0]] = -1;
            }
            pageindex[0] = pageDir[sym][pos[BK]][pos[WK]].index;
            pageIndexTable[pageindex[0]] = 0;
            fseeko(fp, pageindex[0]*pagesize, SEEK_SET);
            fread(addr, 1, pagesize, fp);
            return addr;
        case nine:
            int pagenr;
            if ((pagenr=pageIndexTable[pageDir[sym][pos[BK]][pos[WK]].index]) >= 0) {   //is the page in memory?
                pageage[pagenr] = date;
                return addr + pagenr*pagesize;
            }
            int oldest = 0;             //we need to load a page from disk
            unsigned oldestdate = UINT_MAX;    //find oldest page in memory
            for (unsigned x=0; x<9; x++) 
                if (pageindex[x] >= 0) {
                    if (pageage[x] < oldestdate) {
                        oldestdate = pageage[x];
                        oldest = x;
                    }
                } else {
                    oldest = x;         //good. unused page found, use it.
                    break;
                }
             
            if (pageindex[oldest] >= 0) { //if the page was used, we need to
                if (!readonly) {
                    fseeko(fp, pageindex[oldest]*pagesize, SEEK_SET);
                    fwrite(addr + pagesize*oldest, 1, pagesize, fp);    //write is back
   //                 printf("swap out file %s, page %d at index %d\n", name, pageindex[oldest], oldest);
                }
                pageIndexTable[pageindex[oldest]] = -1;    //and invalidate the entry in the index table pointing to it
            }
            pageindex[oldest] = pageDir[sym][pos[BK]][pos[WK]].index;
            pageage[oldest] = date++;
            pageIndexTable[pageindex[oldest]] = oldest;
            fseeko(fp, pageindex[oldest]*pagesize, SEEK_SET);
            fread(addr + pagesize*oldest, 1, pagesize, fp);
 //           printf("swap in file %s, page %d at index %d\n", name, pageindex[oldest], oldest);
            return addr + pagesize*oldest;
        }           
    }

    bool set(unsigned pos[]) {
        char *p = byteaddr(pos);
        unsigned bitadr = getbitadr(pageDir[sym][pos[BK]][pos[WK]].symops, pos);
        bool oldbit = (p[bitadr >> 3] & (1 << (bitadr & 7))) != 0;
        p[bitadr >> 3] |= 1 << (bitadr & 7);
        if (diagonal[pos[WK]] && diagonal[pos[BK]] && sym) {
            bitadr = getbitadr(pageDir[sym][pos[BK]][pos[WK]].symops ^ DFLIP, pos);
            p[bitadr >> 3] |= 1 << (bitadr & 7);
        }
        return oldbit;
    }

    bool get(unsigned pos[]) {
        char *p = byteaddr(pos);
        unsigned bitadr = getbitadr(pageDir[sym][pos[BK]][pos[WK]].symops, pos);
        return (p[bitadr >> 3] & (1 << (bitadr & 7))) != 0;
    }

};

class Endgame {
    unsigned cnt;                    //number of pieces
    unsigned pos[MAXPIECES], currPage;        //position of piece i
    int     type[MAXPIECES];        //type of piece i
    int     sametype[MAXPIECES];    //sametype[i] is true, if i and i+1 are of the same type
    Endgame *subeg[MAXPIECES][5];   //[0] = captured, [1]-[4] promotions
    int     board[0x10][0x10];
    str     name;
    bool    stalemate;
    bool    sym;                    //true if no pawn is involved, in this case we got only x-mirroring
    bool    b_endpos;
    bool    trivial;                //true if endgame has no winnig positions
    bool    promoiter;               //true if only positions with pawns on promotion squares should be iterated
    unsigned long long count;
    str fnamewtmp, fnamebtmp, fnameweg, fnamebeg, fnamewd, fnamebd;
    int wdfile, bdfile;

    int     solsrc, soldst;
    //matew - white to move, 1 if white mates in ply moves
    //mateb - black to move, 1 if black is mated/gets mated in ply moves

public:
    Swapfile   matew, mateb, neww, newb;

    void startpos() {
        currPage = 0;
        pos[WK] = posTable[sym][currPage].wk;
        board[0][pos[WK]] = type[WK];
        pos[BK] = posTable[sym][currPage].bk;
        board[0][pos[BK]] = type[BK];
        int p = 0;
        for (unsigned i=2; i<cnt; i++) {
            unsigned pawn;
            if (promoiter)
                pawn = type[i] == 5 ? 0x70 : 0x00;
            else
                pawn = abs(type[i]) == 5 ? 0x10 : 0x00;
            while (board[0][p+pawn] && !promoiter)
                p++;
            pos[i] = p+pawn;
            board[0][p+pawn] = type[i];
        }
        b_endpos = false;
	for (unsigned i=2; i<cnt; i++) 
	    if (sametype[i])
		for (unsigned j=i+sametype[i]; j>i; j--)
		    if (pos[j] > pos[i]) {
			int temp = pos[i];
			pos[i] = pos[j];
			pos[j] = temp;
		    }
			
    }

    void nextpos(int n) {
        if (n>1) {
            board[0][pos[n]] = 0;

            do {
                pos[n]++;
                if ((pos[n] & 8) || (pos[n] > pos[n-1] && sametype[n-1])) {
                    pos[n]+=8;
                    
                    if ((pos[n] >= 0x80) || 
                        (pos[n] >= 0x70 && ((type[n] == -5 || (type[n] == 5 && !promoiter)))) ||
                        (pos[n] > pos[n-1] && sametype[n-1])) {
                        nextpos(n-1);
                        if (promoiter)
                            pos[n] = type[n] == 5 ? 0x70 : 0x00;
                        else
                            pos[n] = abs(type[n]) == 5 ? 0x10 : 0x00;
                        if (b_endpos)
                            return;
                
                    }
                }
            } while (board[0][pos[n]] && !promoiter);

            board[0][pos[n]] = type[n];

        } else {
            board[0][pos[WK]] = 0;
            board[0][pos[BK]] = 0;
            currPage++;
            if (opts.verbose) {
                printf("\015%5.1f%% %lld", double(currPage)/double(pagecnt[sym])*100.0, count);
                fflush(stdout);
            }
            if (currPage >= pagecnt[sym]) {
                b_endpos = true;
                if (opts.verbose)
                    printf("\n");
            } else {
                pos[WK] = posTable[sym][currPage].wk;
                pos[BK] = posTable[sym][currPage].bk;
                board[0][pos[WK]] = type[WK];
                board[0][pos[BK]] = type[BK];
            }
        }

    }

    void nextpos() {
        nextpos(cnt-1);
    }

    bool endpos() {
        return b_endpos || trivial;
    }

    void setd(int dfile, int ply) {
        if (opts.dtable) {
            short d = ply;
            lseek(dfile, (pageDir[sym][pos[BK]][pos[WK]].index*matew.pagesize*8 + matew.getbitadr(0, pos))*2, SEEK_SET);
            write(dfile, &d, 2);
        }
    }
    
    int checkw() {
        int d;

        if (DIST(KING, pos[WK] - pos[BK]))
            return true;

        for (unsigned p = 2; p<cnt; p++)
            if (type[p] > 0 && board[0][pos[p]] == type[p])
                if ((d = DIST(type[p], pos[p] - pos[BK]))) {
                    unsigned scanpos = pos[BK] + d;
                    bool blocked = false;
                    while (scanpos != pos[p]) {
                        if (board[0][scanpos]) {
                            blocked = true;
                            break;
                        }
                        scanpos += d;
                    }
                    if (!blocked)
                        return true;
                }
        return false;
    }

    //returns true if black checks
    int checkb() {
        int d;

        if (DIST(KING, pos[WK] - pos[BK]))
            return true;

        for (unsigned p = 2; p<cnt; p++)
            if (type[p] < 0 && board[0][pos[p]] == type[p])
                if ((d = DIST(type[p], pos[p] - pos[WK]))) {
                    unsigned scanpos = pos[WK] + d;
                    bool blocked = false;
                    while (scanpos != pos[p]) {
                        if (board[0][scanpos]) {
                            blocked = true;
                            break;
                        }
                        scanpos += d;
                    }
                    if (!blocked)
                        return true;
                }
        return false;
    }

    //test position, in which piece p is moved from src to dst.
    //if the resulting position is illegal, return false.
    //if the resulting position has a tablevalue == bit,
    //return true, else false.
    bool testw(Swapfile &t, int p, int src, int dst, bool bit) {
        bool ret;
        int oldp = board[0][dst];
        board[0][src] = 0;
        board[0][dst] = type[p];
        pos[p] = dst;
        if (checkb() || oldp > 0)
            ret = false;
        else {
            ret = t.get(pos) == bit;
            stalemate = false;
        }
        board[0][dst] = oldp;
        board[0][src] = type[p];
        pos[p] = src;
        return ret;
    }

    bool testb(Swapfile &t, int p, int src, int dst, bool bit) {
        bool ret;
        int oldp = board[0][dst];
        board[0][src] = 0;
        board[0][dst] = type[p];
        pos[p] = dst;
        if (checkw() || oldp < 0)
            ret = false;
        else {
            ret = t.get(pos) == bit;
            stalemate = false;
        }
        board[0][dst] = oldp;
        board[0][src] = type[p];
        pos[p] = src;
        return ret;
    }

    int check(int color, int p, int src, int dst) {
        int ret;
        board[0][src] = 0;
        board[0][dst] = type[p];
        if (color == WHITE)
            ret=checkw();
        else
            ret=checkb();
        board[0][dst] = 0;
        board[0][src] = type[p];
        return ret;
    }        
    
    void retroMove(Swapfile &t, int color) {
        int d, oldpos;
            
        for (unsigned p=0; p<cnt; p++)
            switch (type[p] * color) {
            case KING:
                for (d = 0; d < 8; d ++) {
                    oldpos = pos[p];
                    if (!((pos[p]+=dirtab[d]) & 0x88))
                        if (board[0][pos[p]] == 0)
                            if (!check(color, p, oldpos, pos[p]))
                                t.set(pos);
                    pos[p] = oldpos;
                }
                break;
            case KNIGHT:
                for (d = 0; d < 8; d ++) {
                    oldpos = pos[p];
                    if (!((pos[p]+=kdirtab[d]) & 0x88))
                        if (board[0][pos[p]] == 0)
                            if (!check(color, p, oldpos, pos[p]))
                                t.set(pos);
                    pos[p] = oldpos;
                }
                break;
            case QUEEN:
                for (d = 0; d < 8; d ++) {
                    oldpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        pos[p] += dirtab[d];
                        if (pos[p] & 0x88) break;
                        if (board[0][pos[p]] != 0) break;
                        if (!check(color, p, oldpos, pos[p])) 
                        t.set(pos);
                    }
                    pos[p] = oldpos;
                }
                break;
            case BISHOP:
                for (d = 1; d < 8; d+=2) {
                    oldpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        pos[p] += dirtab[d];
                        if (pos[p] & 0x88) break;
                        if (board[0][pos[p]] != 0) break;
                        if (!check(color, p, oldpos, pos[p])) 
                        t.set(pos);
                    }
                    pos[p] = oldpos;
                }
                break;
            case ROOK:
                for (d = 0; d < 8; d+=2) {
                    oldpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        pos[p] += dirtab[d];
                        if (pos[p] & 0x88) break;
                        if (board[0][pos[p]] != 0) break;
                        if (!check(color, p, oldpos, pos[p])) 
                            t.set(pos);
                    }
                    pos[p] = oldpos;
                }
                break;
            case PAWN:
                oldpos = pos[p];
                if (color > 0) {
                    if (oldpos >= 0x20) {
                        pos[p] -= dirtab[2];
                        if (board[0][pos[p]] == 0) {
                            if (!check(color, p, oldpos, pos[p])) 
                                t.set(pos);         
                            if (oldpos < 0x40)    
                                pos[p] -= dirtab[2];
                                if (board[0][pos[p]] == 0) 
                                    if (!check(color, p, oldpos, pos[p])) 
                                        t.set(pos);
                        }
                    }                  
                } else {
                    if (oldpos < 0x60) {
                        pos[p] -= dirtab[6];
                        if (board[0][pos[p]] == 0) {
                            if (!check(color, p, oldpos, pos[p])) 
                                t.set(pos);         
                            if (oldpos >= 0x40)    
                                pos[p] -= dirtab[6];
                                if (board[0][pos[p]] == 0) 
                                    if (!check(color, p, oldpos, pos[p])) 
                                        t.set(pos);
                        }
                    }                  
                }
                pos[p] = oldpos;
            }
        return;
    }

    //if there is at least one legal move with tablevalue == bit,
    //return true, else false.
    bool findMovew(Swapfile &t, int bit) {
        int d, newpos;
        stalemate = true;
        for (unsigned p=0; p<cnt; p++)
            switch (type[p]) {
            case KING:
                for (d = 0; d < 8; d ++)
                    if (!((newpos = pos[p]+dirtab[d]) & 0x88))
                        if (board[0][newpos] <= 0)
                            if (testw(t, p, pos[p], newpos, bit)) {
                                solsrc = pos[p];
                                soldst = newpos;
                                return true;
                            }
                break;
            case KNIGHT:
                for (d = 0; d < 8; d ++)
                    if (!((newpos = pos[p]+kdirtab[d]) & 0x88))
                        if (board[0][newpos] <= 0)
                            if (testw(t, p, pos[p], newpos, bit))
                                return true;
                break;
            case QUEEN:
                for (d = 0; d < 8; d ++) {
                    newpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        newpos += dirtab[d];
                        if (newpos & 0x88)
                            break;
                        if (board[0][newpos] > 0)
                            break;
                        if (testw(t, p, pos[p], newpos, bit)) {
                                solsrc = pos[p];
                                soldst = newpos;
                            return true;
                        }
                        if (board[0][newpos])
                            break;
                    }
                }
                break;
            case BISHOP:
                for (d = 1; d < 8; d+=2) {
                    newpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        newpos += dirtab[d];
                        if (newpos & 0x88)
                            break;
                        if (board[0][newpos] > 0)
                            break;
                        if (testw(t, p, pos[p], newpos, bit))
                            return true;
                        if (board[0][newpos])
                            break;
                    }
                }
                break;
            case ROOK:
                for (d = 0; d < 8; d+=2) {
                    newpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        newpos += dirtab[d];
                        if (newpos & 0x88)
                            break;
                        if (board[0][newpos] > 0)
                            break;
                        if (testw(t, p, pos[p], newpos, bit))
                            return true;
                        if (board[0][newpos])
                            break;
                    }
                }
                break;
            case PAWN:
                newpos = pos[p] + dirtab[2];
                if (board[0][newpos] == 0) {
                    if (testw(t, p, pos[p], newpos, bit)) {
                                                                                solsrc = pos[p];
                                soldst = newpos;

                        return true;
                    } else if (pos[p] <= 0x17) {
                        newpos = pos[p] + 2*dirtab[2];
                        if (board[0][newpos] == 0)
                            if (testw(t, p, pos[p], newpos, bit)) {
                                                            solsrc = pos[p];
                                soldst = newpos;

                                return true;
                            }
                    }
                }
                newpos = pos[p] + dirtab[1];
                if (!(newpos & 0x88))
                    if (board[0][newpos] < 0)
                        if (testw(t, p, pos[p], newpos, bit))
                            return true;
                newpos = pos[p] + dirtab[3];
                if (!(newpos & 0x88))
                    if (board[0][newpos] < 0)
                        if (testw(t, p, pos[p], newpos, bit))
                            return true;
            }
        return false;
    }

    //if there is at least one legal move with tablevalue == bit,
    //return true, else false.
    bool findMoveb(Swapfile &t, int bit) {
        int d, newpos;
        stalemate = true;
        for (unsigned p=0; p<cnt; p++)
            switch (type[p]) {
            case -KING:
                for (d = 0; d < 8; d ++)
                    if (!((newpos = pos[p]+dirtab[d]) & 0x88))
                        if (board[0][newpos] >= 0)
                            if (testb(t, p, pos[p], newpos, bit))
                                return true;
                break;
            case -KNIGHT:
                for (d = 0; d < 8; d ++)
                    if (!((newpos = pos[p]+kdirtab[d]) & 0x88))
                        if (board[0][newpos] >= 0)
                            if (testb(t, p, pos[p], newpos, bit))
                                return true;
                break;
            case -QUEEN:
                for (d = 0; d < 8; d ++) {
                    newpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        newpos += dirtab[d];
                        if (newpos & 0x88)
                            break;
                        if (board[0][newpos] < 0)
                            break;
                        if (testb(t, p, pos[p], newpos, bit))
                            return true;
                        if (board[0][newpos])
                            break;
                    }
                }
                break;
            case -BISHOP:
                for (d = 1; d < 8; d+=2) {
                    newpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        newpos += dirtab[d];
                        if (newpos & 0x88)
                            break;
                        if (board[0][newpos] < 0)
                            break;
                        if (testb(t, p, pos[p], newpos, bit))
                            return true;
                        if (board[0][newpos])
                            break;
                    }
                }
                break;
            case -ROOK:
                for (d = 0; d < 8; d+=2) {
                    newpos = pos[p];
                    for (int l = 1; l<8; l++) {
                        newpos += dirtab[d];
                        if (newpos & 0x88)
                            break;
                        if (board[0][newpos] < 0)
                            break;
                        if (testb(t, p, pos[p], newpos, bit))
                            return true;
                        if (board[0][newpos])
                            break;
                    }
                }
                break;
            case -PAWN:
                newpos = pos[p] + dirtab[6];
                if (board[0][newpos] == 0)
                    if (testw(t, p, pos[p], newpos, bit))
                        return true;
                    else if (pos[p] >= 0x60) {
                        newpos = pos[p] + 2*dirtab[6];
                        if (board[0][newpos] == 0)
                            if (testw(t, p, pos[p], newpos, bit))
                                return true;
                    }
                newpos = pos[p] + dirtab[5];
                if (!(newpos & 0x88))
                    if (board[0][newpos] < 0)
                        if (testw(t, p, pos[p], newpos, bit))
                            return true;
                newpos = pos[p] + dirtab[7];
                if (!(newpos & 0x88))
                    if (board[0][newpos] < 0)
                        if (testw(t, p, pos[p], newpos, bit))
                            return true;
            }
        return false;
    }

    void niceout(unsigned long long n1, unsigned long long n2, unsigned long long tot) {
        printf("%13llu %6.2f %13llu %6.2f %13llu %6.2f\n", n1, (double)n1/tot*100.0, n2, (double)n2/tot*100.0, n1+n2, double(n1+n2)/tot *100.0);
    }

    Endgame *findEndgame(char *t) {
        for (Endgame **eg = egCollection; eg < egEnd; eg++)
            if (!strcmp((*eg)->name, t))
                return *eg;

        Endgame *temp = new Endgame(t);
        *egEnd++ = temp;
        return temp;
    }

    Endgame() {
        name[0] = 0;
    }

    ~Endgame() {
        matew.free();
        mateb.free();
    }   

    Endgame(const char *sig) {
        strcpy(name, sig);
        sort(name);
        //if we have no white pieces, the endgame is trivial, since all
        //positions are not winning and the whole table is filled with '0'.
        //just make sure, endpos() always returns true.
        if (!strcmp(name, "B") || !strcmp(name, "N") || !strlen(name) || name[0] > 'Z') {
            printf("Not building %s, because it is trivial\n", name);
            trivial = true;
            return;
        }
        
        trivial = false;
        printf("Building %s\n", name);
        
        sym = true;
        str s;
        strcpy(s, "Kk");
        strcat(s, name);
        cnt = strlen(s);
        type[0] = KING;
        type[1] = -KING;

        for (unsigned i = 2; i<cnt; i++) {
            switch (s[i]) {
            case 'Q':
                type[i] = QUEEN;
                break;
            case 'q':
                type[i] = -QUEEN;
                break;
            case 'B':
                type[i] = BISHOP;
                break;
            case 'b':
                type[i] = -BISHOP;
                break;
            case 'R':
                type[i] = ROOK;
                break;
            case 'r':
                type[i] = -ROOK;
                break;
            case 'N':
                type[i] = KNIGHT;
                break;
            case 'n':
                type[i] = -KNIGHT;
                break;
            case 'P':
                type[i] = PAWN;
                s[i] = 'Q';
                subeg[i][1] = findEndgame(s+2);
                s[i] = 'R';
                subeg[i][2] = findEndgame(s+2);
                s[i] = 'N';
                subeg[i][3] = findEndgame(s+2);
                s[i] = 'B';
                subeg[i][4] = findEndgame(s+2);
                s[i] = 'P';
                sym = false;
                break;
            case 'p':
                type[i] = -PAWN;
                s[i] = 'q';
                subeg[i][1] = findEndgame(s+2);
                s[i] = 'r';
                subeg[i][2] = findEndgame(s+2);
                s[i] = 'n';
                subeg[i][3] = findEndgame(s+2);
                s[i] = 'b';
                subeg[i][4] = findEndgame(s+2);
                s[i] = 'p';
                sym = false;
                break;
            default:
                fprintf(stderr, "unknown piecetype %c\n", s[i]);
                exit(1);
            }

            //build subendgames
            //for each piece i (if i is captured)
            char t[256];
            unsigned k=0;
            for (unsigned j=2; j<=cnt; j++)
                if (i!=j)
                    t[k++] = s[j];
                    
            subeg[i][0] = findEndgame(t);
            //check for equal piece types
            sametype[i] = 0;
            for (unsigned j = i+1; j<cnt && s[j] == s[i]; j++)
                sametype[i]++;

        }
        
        if (opts.verbose) printf("Piece types are %s (%d)\n", s, cnt);
        
        unsigned long long posTotal, posIter, posWIll, posBIll, posWLeg, posBLeg, posLeg, posWin, posWCap, posBCap, posWNWCap, posBNWCap;
        posTotal = posIter = posWIll = posBIll = posWLeg = posBLeg = posLeg = posWin = posWCap = posBCap = posWNWCap = posBNWCap = 0;
        unsigned long long posMate[300];

        strcpy(fnamewtmp, name);
        strcpy(fnamebtmp, name);
        strcpy(fnameweg, name);
        strcpy(fnamebeg, name);
        strcpy(fnamewd, name);
        strcpy(fnamebd, name);
        strcat(fnamewtmp, ".w.tmp");
        strcat(fnamebtmp, ".b.tmp");
        strcat(fnameweg, ".w.eg");
        strcat(fnamebeg, ".b.eg");
        strcat(fnamewd, ".w.d.eg");
        strcat(fnamebd, ".b.d.eg");
        
        unlink(fnamewtmp);
        unlink(fnamebtmp);
        
        if (!opts.newstart) {
            struct stat buf;
            if (stat(fnameweg, &buf)==0 && stat(fnamebeg, &buf)==0) {
                if (opts.verbose)
                    printf("found precalculated endgame %s\n", name);
                matew.alloc(fnameweg, true, one, sametype, cnt, sym);
                mateb.alloc(fnamebeg, true, one, sametype, cnt, sym);
                return;
            }
        }
            
        if (opts.usepaging) {
            matew.alloc(fnamewtmp, false, one, sametype, cnt, sym);
            mateb.alloc(fnamebtmp, false, one, sametype, cnt, sym);
        } else {
            matew.alloc(fnamewtmp, false, full, sametype, cnt, sym);
            mateb.alloc(fnamebtmp, false, full, sametype, cnt, sym);
        }

        if (opts.dtable) {
            wdfile = open(fnamewd, O_RDWR|O_CREAT|O_TRUNC, 0777);
            bdfile = open(fnamebd, O_RDWR|O_CREAT|O_TRUNC, 0777);
        }
        
        printf("%1$d pages with %2$llu (0x%2$llx) bytes\n", pagecnt[sym], matew.pagesize);

        count = 0;
        if (opts.verbose) printf("Counting illegal positions white\n");
        ITERATE(this) {
            if (checkw()) {
                matew.set(pos);
                count++;
            }
        }
        posWIll = count;

        count = 0;
        if (opts.verbose) printf("Counting illegal positions black\n");
        ITERATE(this)
            if (checkb()) {
                mateb.set(pos);
                count++;
            }
        posBIll = count;
	
        posWin = 0;
        for (unsigned capturing = 0; capturing < cnt; capturing++)
            if (type[capturing] < 0)
                for (unsigned cap = 2; cap < cnt; cap += 1 + sametype[cap]) {
                    if (cap == capturing || type[cap] < 0)
                        continue;
                    subeg[cap][0]->count = 0;
                    if (opts.verbose) printf("Transfering sub endgame %s (%cx%c)\n", subeg[cap][0]->name, s[capturing], s[cap]);
                    ITERATE(subeg[cap][0])
                        if (subeg[cap][0]->matew.get(subeg[cap][0]->pos) ) {
                            int j=0;
                            for (unsigned i=0; i<cnt; i++) {
                                if (i == cap)
                                    pos[i] = pos[capturing];
                                else
                                    pos[i] = subeg[cap][0]->pos[j++];
                            }
                            if (!matew.set(pos)) {
                                setd(wdfile, 1);
                                subeg[cap][0]->count++;
                                posBCap++;
                            }
                        }
                        else
                            posBNWCap++;
                }
            else
                for (unsigned cap = 2; cap < cnt; cap += 1 + sametype[cap]) {
                    if (cap == capturing || type[cap] > 0)
                        continue;
                    subeg[cap][0]->count = 0;
                    if (opts.verbose) printf("Transfering sub endgame %s (%cx%c)\n", subeg[cap][0]->name, s[capturing], s[cap]);
                    ITERATE(subeg[cap][0])
                        if (subeg[cap][0]->mateb.get(subeg[cap][0]->pos) ) {
                            int j=0;
                            for (unsigned i=0; i<cnt; i++) {
                                if (i == cap)
                                    pos[i] = pos[capturing];
                                else
                                    pos[i] = subeg[cap][0]->pos[j++];
                            }
                            if (!mateb.set(pos)) {
                                setd(bdfile, 1);
                                subeg[cap][0]->count++;
                                posWCap++;
			    }
                        }
                        else
                            posWNWCap++;
                    }
                
        promoiter = true;
        for (unsigned promo = 2; promo < cnt; promo++) 
            switch (type[promo]) {
            case PAWN:
                for (int i=1; i<=4; i++)
                    if (opts.verbose) if (!subeg[promo][i]->trivial) printf("Transfering sub endgame %s (promotion)\n", subeg[promo][i]->name);
                count = 0;
                ITERATE(this)
                    for (int i=1; i<=4; i++) {
                        if (!subeg[promo][i]->trivial) {
                            for (unsigned j=0; j<cnt; j++)
                                subeg[promo][i]->pos[j]=pos[j];
                            if (subeg[promo][i]->mateb.get(subeg[promo][i]->pos)) {
                                setd(bdfile, 1);
                                mateb.set(pos);
                                count++;                                
                                break;
                            }
                        }
                    }
                break;
            case -PAWN:
                for (int i=1; i<=4; i++)
                    if (opts.verbose) if (!subeg[promo][i]->trivial) printf("Transfering sub endgame %s (promotion)\n", subeg[promo][i]->name);
                count = 0;
                ITERATE(this) {
                    bool found=false;
                    for (int i=1; i<=4; i++) 
                        if (!subeg[promo][i]->trivial) {
                            for (unsigned j=0; j<cnt; j++)
                                subeg[promo][i]->pos[j]=pos[j];
                            if (subeg[promo][i]->matew.get(subeg[promo][i]->pos)) {
                                found = true;
                                count++;                                
                                break;
                            }
                            if (!found) {
                                matew.set(pos);
                                setd(wdfile, 1);
                            }
                        }
                    }
                break;
            }

        promoiter = false;
        
        posTotal = matew.pagesize*pagecnt[sym]*8;
        count = 0;
        if (opts.verbose) printf("Counting iterated positions\n");
        ITERATE(this)
            count++;
        posIter = count;
        if (opts.verbose) printf("%llu iterated positions\n", posIter);
            
        posWLeg = posIter - posWIll;
        posBLeg = posIter - posBIll;
        posLeg = posBLeg + posWLeg;
        int ply=0;
        do {
            count = 0;
            if (opts.verbose) printf("Black is mated\n");
            if (opts.usepaging) {
//                matew.realloc(false, nine); 
                matew.realloc(false, one); 
            } 
            ITERATE(this)
                if (!mateb.get(pos) && checkw() && !findMoveb(matew, false)) {
                    mateb.set(pos);
//                    printf("%2x %2x %2x %2x:%2x %2x\n", pos[0], pos[1], pos[2], pos[3], solsrc, soldst);
                    setd(bdfile, ply);
                    count++;
                }
            
            posMate[ply++] = count;
            posWin += count;
            
            unlink("neww");
            unlink("newb");

            count = 0;
            if (opts.verbose) printf("White wins in%5d\n",(ply+1)/2);
            if (opts.usepaging) {
//                newb.alloc("newb", false, nine, sametype, cnt, sym);  //only opened to be closed in next block
                newb.alloc("newb", false, one, sametype, cnt, sym);  //only opened to be closed in next block
                mateb.realloc(true, one); //used in findMove, local paging needed
//                matew.realloc(false, nine);  //linear reading
                matew.realloc(false, one);  //linear reading
            } else
                newb.alloc("newb", false, full, sametype,cnt, sym);  //get whole file
            neww.alloc("neww", false, one, sametype,cnt, sym);
            ITERATE(this)
                if (!matew.get(pos) && findMovew(mateb, true)) {
                    matew.set(pos);
                    retroMove(newb, BLACK);
                    setd(wdfile, ply);
                    count++;
                }
            if (!count) break;

            posMate[ply++] = count;
            posWin += count;

            for ( ; ; ) {
                count = 0;
                if (opts.verbose) printf("Black loses in%5d\n", ply/2);
                if (opts.usepaging) {
//                    neww.alloc(false, nine);  //only opened to be closed in next block
                    neww.alloc(false, one);  //only opened to be closed in next block
                    newb.realloc(true, one);
//                    matew.realloc(true, nine); 
                    matew.realloc(true, one); 
                    mateb.realloc(false, one);
                } else 
                    neww.alloc(false, full);  //only opened to be closed in next block
                ITERATE(this)
                    if (newb.get(pos) && !mateb.get(pos) && !findMoveb(matew, false) && !stalemate) {
                        mateb.set(pos);
//                        if (pos[0] == 0 && pos[1] == 0x22 && pos[2] == 0x15)
//                            printf("*******************\n");
                        setd(bdfile, ply);
                        retroMove(neww, WHITE);
                        count++;
                    }
                if (!count) break;
                newb.free();
                unlink("newb");

                posMate[ply++] = count;
                posWin += count;
            
                count = 0;
                if (opts.verbose) printf("White wins  in%5d\n", (ply+1)/2);
                if (opts.usepaging) {
//                    newb.alloc(false, nine);  //used in retromove, local paging
                    newb.alloc(false, one);  //used in retromove, local paging
                    neww.realloc(true, one);    //used in linear reading
//                    mateb.realloc(true, nine); //used in findMove, local paging needed
                    mateb.realloc(true, one); //used in findMove, local paging needed
                    matew.realloc(false, one);  //linear reading
                } else
                    newb.alloc(false, full);  //get whole file
                ITERATE(this) {
                    
                    if (neww.get(pos) && !matew.get(pos) /*&& findMovew(mateb, true)*/) {
//                        if (ply==37)
//                                            printf("%2x %2x %2x\n", pos[0], pos[1], pos[2]);
//                        if (!get(neww))
//                            printf("%2x %2x %2x %2x:%2x %2x\n", pos[0], pos[1], pos[2], pos[3], solsrc, soldst);
                        matew.set(pos);
                        setd(wdfile, ply);
                        retroMove(newb, BLACK);
                        count++;
                    }
                }    
                if (!count) break;
                neww.free();
                unlink("neww");
        
                posMate[ply++] = count;
                posWin += count;

            }
        } while(0);

//	count = 0;
//	ITERATE(this) {
//	    if (!matew.get(pos)) {
//    		count++;
//		if (count % 1000 == 0)
//		    printf(" %2x %2x %2x %2x\n", pos[0], pos[1], pos[2], pos[3]);
//	    }
//di	}
        for(;0;) {
            scanf("%i%i%i",&pos[0],&pos[1],&pos[2]);
            printf("wtm: %d\nbtm: %d\n", matew.get(pos), mateb.get(pos));
        }
        neww.free();
        newb.free();
        unlink("neww");
        unlink("newb");
        rename(fnamewtmp, fnameweg);
        rename(fnamebtmp, fnamebeg);

        if (opts.verbose) {
            printf("---Statistics for %s:\n", name);
            printf("total positions per color: %24lld\n", posTotal);
            printf("winning captures:"); niceout(posWCap, posBCap, posTotal*2);
            printf("not win captures:"); niceout(posWNWCap, posBNWCap, posTotal*2);
            printf("iterated pos:    "); niceout(posIter, posIter, posTotal*2);
            printf("---of iterated\n");
            printf("illegal pos:     "); niceout(posWIll, posBIll, posIter*2);
            printf("legal positions: "); niceout(posWLeg, posBLeg, posIter*2);
            printf("---of legal\n");
            printf("win/not winning: "); niceout(posWin, posLeg-posWin, posLeg);
            for (int i = 0; i<ply; i+=2) {
                printf("winning in %4d: ", i/2);
                niceout(i==0?0:posMate[i-1], posMate[i], posLeg);
            }
        }
    }
};

int mirror(unsigned p, unsigned flags) {
    unsigned x = p & 7;
    unsigned y = p >> 4;
    if (flags & XFLIP)
        x ^= 7;
    if (flags & YFLIP)
        y ^= 7;
    if (flags & DFLIP) {
        int temp = x;
        x = y;
        y = temp;
    }
    return x+8*y;
}

void sort(char *s) {
    for (unsigned i = 0; i<strlen(s); i++)
        for (unsigned j = 0; j<strlen(s)-1; j++)
            if (s[j] > s[j+1]) {
                char temp = s[j];
                s[j] = s[j+1];
                s[j+1] = temp;
            }
}

void init() {
    for (int d=1; d<8; d++) {
        for (int dir = 0; dir < 8; dir++) {
            DIST(QUEEN, dirtab[dir] * d) = dirtab[dir];
            DIST(-QUEEN, dirtab[dir] * d) = dirtab[dir];
        }
        for (int dir = 0; dir < 8; dir+=2) {
            DIST(ROOK, dirtab[dir] * d) = dirtab[dir];
            DIST(-ROOK, dirtab[dir] * d) = dirtab[dir];
        }
        for (int dir = 1; dir < 8; dir+=2) {
            DIST(BISHOP, dirtab[dir] * d) = dirtab[dir];
            DIST(-BISHOP, dirtab[dir] * d) = dirtab[dir];
        }
    }
    for (int dir = 0; dir < 8; dir++) {
        DIST(KING, dirtab[dir] ) = dirtab[dir];
        DIST(-KING, dirtab[dir] ) = dirtab[dir];
    }

    DIST(PAWN, dirtab[5]) = dirtab[5];
    DIST(PAWN, dirtab[7]) = dirtab[7];

    DIST(-PAWN, dirtab[1]) = dirtab[1];
    DIST(-PAWN, dirtab[3]) = dirtab[3];

    for (int dir = 0; dir < 8; dir++) {
        DIST(KNIGHT, kdirtab[dir]) = kdirtab[dir];
        DIST(-KNIGHT, kdirtab[dir]) = kdirtab[dir];
    }

    for (int p=0; p<0x80; p++)
        for (int flags = 0; flags < 8; flags++)
            mirrortab[p][flags] = mirror(p, flags);

    for (int p=0; p<=0x77; p+=0x11)
        diagonal[p] = true;
    
    memset(pageDir, -1, sizeof(pageDir));   
    for (int sym = 0; sym < 2; sym++) {
        unsigned currpage = 0;
        for (int wky = 0; wky<8; wky++)
            for (int wkx = 0; wkx<8; wkx++)
                for (int bky = 0; bky<8; bky++)
                    for (int bkx = 0; bkx<8; bkx++)
                        if (abs(wkx-bkx) > 1 || abs(wky-bky) > 1) {
                            int flags = 0;
                            int mwkx = wkx;
                            int mbkx = bkx;
                            int mwky = wky;
                            int mbky = bky;
                            if (wkx>3) {
                                mwkx = 7-wkx;
                                mbkx = 7-bkx;
                                flags |= XFLIP;
                            }
                            if (wky>3 && sym) {
                                mwky = 7-wky;
                                mbky = 7-bky;
                                flags |= YFLIP;
                            }

                            if ((mwky > mwkx || (mwkx == mwky && mbky > mbkx)) && sym) {
                                flags |= DFLIP;
                                int temp = mwkx;
                                mwkx = mwky;
                                mwky = temp;
                                temp = mbkx;
                                mbkx = mbky;
                                mbky = temp;
                            }

                            int wk = 0x10*wky+wkx;
                            int bk = 0x10*bky+bkx;
                            int mwk = 0x10*mwky+mwkx;
                            int mbk = 0x10*mbky+mbkx;
                            pageDir[sym][bk][wk].symops = flags;
                            if (!flags) {
                                pageDir[sym][bk][wk].index = currpage;
                                posTable[sym][currpage].bk = bk;
                                posTable[sym][currpage].wk = wk;
                                currpage++;
                            } else {
                                pageDir[sym][bk][wk].index = pageDir[sym][mbk][mwk].index;
                            }

                        }
        pagecnt[sym] = currpage;
    }

}

int main(int argc, char *argv[]) {
    char opt;
    char s1[256], s2[256];
    
    if (argc == 1) {
        exit(1);
    }

    while ((opt = getopt(argc, argv, "hcnvpdr")) > 0) {
        switch(opt) {
        case 'h':
            printf("Usage: endgame [options] piecepattern\n\
                    where piecepattern is [BbNnPpQrRr]+ and options are\n\
                    -h: this text\n\
                    -s: show various statistics for a already calculated endgame\n\
                    -n: start new from scratch, overwriting old files\n\
                    -v: be verbose about what is done\n\
                    -p: use own swapping code\n\
                    -r: produce reverse table\n\
                    -d: produce distance-to-win table\n");
            exit(0);
            break;
        case 'n':
            opts.newstart = true;
            break;
        case 'v':
            opts.verbose = true;
            break;
        case 'p':
            opts.usepaging = true;
            break;
        case 'd':
            opts.dtable = true;
            break;
        case 'r':
            opts.rtable = true;
            break;
        }
    }

    int l=strlen(argv[optind]);

    if (l+2 > MAXPIECES) {
        fprintf(stderr, "Maximum %d pieces including both kings", MAXPIECES);
        exit(1);
    }

    init();

    unsigned i;
    for (i=0; i<strlen(argv[optind]); i++) {
        s1[i] = argv[optind][i];
        s2[i] = argv[optind][i] ^ 0x20;
    }
    s1[i] = 0;
    s2[i] = 0;
    
    sort(s1);
    sort(s2);
    
    egEnd = egCollection;
    Endgame e1(s1);
    if (strcmp(s1, s2) && opts.rtable) {
        Endgame e2(s2);
    }
    
    for (Endgame **eg = egCollection; eg < egEnd; eg++)
        delete *eg;
        
}

#define MAXENDGAMES 1009

struct EndgameEntry {
    unsigned sig[12];
    Swapfile *bitw, *bitb;
    bool loaded;
    char dtabw[NAME_MAX];
    char dtabb[NAME_MAX];
};

EndgameEntry endgamehash[MAXENDGAMES];

int sel(const struct dirent *dentry) {
    return fnmatch("*.eg", dentry->d_name, 0);
}
    
void endgameinit(char *directory) {
    struct dirent **namelist;
    int n = scandir(directory, &namelist, sel, alphasort);
    if (n>0) {
        while(n--) {
            int baselen = strchr(namelist[n]->d_name, '.') - namelist[n]->d_name;
            str basename, fname;
            strncpy(basename, namelist[n]->d_name, baselen);
            basename[baselen] = 0;
            unsigned cnt = strlen(basename);
            int sametype[MAXPIECES];
            unsigned sym = 1;
            for (unsigned i = 0; i<cnt; i++) {
                if (basename[i] == 'p' || basename[i] == 'P')
                    sym = 0;
                sametype[i] = 0;
                for (unsigned j = i+1; j<cnt && basename[j] == basename[i]; j++)
                    sametype[i]++;
            }
            EndgameEntry h;
            
            sprintf(fname, "%s.w.eg", basename);
            h.bitw = new Swapfile;
            h.bitw->alloc(fname, true, one, sametype, cnt, sym);

            sprintf(fname, "%s.b.eg", basename);
            h.bitb = new Swapfile;
            h.bitb->alloc(fname, true, one, sametype, cnt, sym);

            FILE *f;
            sprintf(fname, "%s.w.d.eg", basename);
            if ( (f=fopen(fname, "r")) ) {
                fclose(f);
                strcpy(h.dtabw, fname);
            } else
                h.dtabw[0] = 0;
                
            sprintf(fname, "%s.b.d.eg", basename);
            if ( (f=fopen(fname, "r")) ) {
                fclose(f);
                strcpy(h.dtabb, fname);
            } else
                h.dtabb[0] = 0;
                
            free(namelist[n]);
        }
        free(namelist);
    } else
        fprintf(stderr, "no endgame tables found in %s\n", directory);
}

bool endgameevalx(int color, const PieceList piecelist[13], float *score) {
    unsigned hash = 0;
    unsigned hash_rev = 0;
    for (int i=-6; i<=6; i++) 
        if (i) {
            hash = (hash<<2) ^ CNT(i);
            hash_rev = (hash_rev<<2) ^ CNT(-i);
        }

    for (int j = WHITE; j >= BLACK; j -= 2, hash = hash_rev)    
        while (endgamehash[hash %= MAXENDGAMES].loaded) {
            unsigned pos[MAXPIECES], *ppos;
            for (int i=-6; i<=6; i++)
                if (i && endgamehash[hash].sig[i+6] != CNT(j*i)) 
                    goto notfound;
            FILE *f;
            pos[WK] = POS(j*KING, 0);
            pos[BK] = POS(j*-KING, 0);
            ppos = pos+2;
            for (int c = 1; c >= -1; c -= 2) {
                for (int i=CNT(j*c*BISHOP); i>0; i--)
                    *ppos++ = POS(j*c*BISHOP, i);
                for (int i=CNT(j*c*KNIGHT); i>0; i--)
                    *ppos++ = POS(j*c*KNIGHT, i);
                for (int i=CNT(j*c*PAWN); i>0; i--)
                    *ppos++ = POS(j*c*PAWN, i);
                for (int i=CNT(j*c*QUEEN); i>0; i--)
                    *ppos++ = POS(j*c*QUEEN, i);
                for (int i=CNT(j*c*ROOK); i>0; i--)
                    *ppos++ = POS(j*c*ROOK, i);
            }
            if (j*color == WHITE) {
                f = fopen(endgamehash[hash].dtabw, "r");
            } else {
                f = fopen(endgamehash[hash].dtabb, "r");
            }
            fseek(f, endgamehash[hash].bitw->linearaddr(pos)*2, SEEK_SET);
            short val;
            fread(&val, 2, 1, f);
            fclose(f);
            if (val > 0)
                *score = j*(-val-MATE_VALUE);
            else if (val < 0)
                *score = j*(-val+MATE_VALUE);
            else
                *score = 0;    
            return true;
        notfound:
            hash++;
        }
    
    return false;
}

bool endgameeval(int color, const PieceList piecelist[13], float *score) {
    unsigned hash = 0;
    unsigned hash_rev = 0;
    for (int i=-6; i<=6; i++) 
        if (i) {
            hash = (hash<<2) ^ CNT(i);
            hash_rev = (hash_rev<<2) ^ CNT(-i);
        }

    while (endgamehash[hash %= MAXENDGAMES].loaded) {
        unsigned pos[MAXPIECES], *ppos;
        for (int i=-6; i<6; i++)
            if (i && endgamehash[hash].sig[i] != CNT(i)) 
                goto notfound;
        pos[WK] = POS(KING, 0);
        pos[BK] = POS(-KING, 0);
        ppos = pos+2;
        for (int c = 1; c >= -1; c -= 2) {
            for (int i=CNT(c*BISHOP); i>0; i--)
                *ppos++ = POS(c*BISHOP, i);
            for (int i=CNT(c*KNIGHT); i>0; i--)
                *ppos++ = POS(c*KNIGHT, i);
            for (int i=CNT(c*PAWN); i>0; i--)
                *ppos++ = POS(c*PAWN, i);
            for (int i=CNT(c*QUEEN); i>0; i--)
                *ppos++ = POS(c*QUEEN, i);
            for (int i=CNT(c*ROOK); i>0; i--)
                *ppos++ = POS(c*ROOK, i);
        }
        if (color == WHITE)
            *score = endgamehash[hash].bitw->get(pos)? -MATE_VALUE:0;
        else
            *score = endgamehash[hash].bitb->get(pos)? -MATE_VALUE:0;
        return true;
    notfound:
        hash++;
    }
    
    while (endgamehash[hash_rev %= MAXENDGAMES].loaded) {
        unsigned pos[MAXPIECES], *ppos;
        for (int i=-6; i<6; i++)
            if (i && endgamehash[hash].sig[i] != CNT(-i)) 
                goto notfound2;
        pos[WK] = POS(-KING, 0);
        pos[BK] = POS(KING, 0);
        ppos = pos+2;
        for (int c = -1; c >= 1; c += 2) {
            for (int i=CNT(c*BISHOP); i>0; i--)
                *ppos++ = POS(c*BISHOP, i);
            for (int i=CNT(c*KNIGHT); i>0; i--)
                *ppos++ = POS(c*KNIGHT, i);
            for (int i=CNT(c*PAWN); i>0; i--)
                *ppos++ = POS(c*PAWN, i);
            for (int i=CNT(c*QUEEN); i>0; i--)
                *ppos++ = POS(c*QUEEN, i);
            for (int i=CNT(c*ROOK); i>0; i--)
                *ppos++ = POS(c*ROOK, i);
        }
        if (color == BLACK)
            *score = endgamehash[hash_rev].bitw->get(pos)? MATE_VALUE:0;
        else
            *score = endgamehash[hash_rev].bitb->get(pos)? MATE_VALUE:0;
        return true;
    notfound2:
        hash_rev++;
    }
    
    return false;
}

