/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#ifndef BOARD_H
#define BOARD_H

#include "globals.h"
#include "hash.h"

#define REPIDX (((b->key.main.key ^ b->color ^ b->epstatus ^ (b->o_o<<9)) ) & (REPTABLESIZE - 1))

#define COL ((signed)sizeof(int))
#define ROW (16*COL)
#define BOARD_SHIFT 9 // log2(COL*ROW)

//since all our positions are differences between absolut addresses (something like char *),
//and not differences between pointers (like int* or piece_t*) we have to convert.
//the purpose is to let the compiler save some shifts.
//BOARD22(x) is almost the same as board[2][2 + x/sizeof(int)], but without the explicit shift right
//( /sizeof ) and the implicit shift left when the compiler hast to multiply (x/sizeof(int)) again
//with sizeof(int) to get the real address.
//these macros assume there is a pointer b to a struct Board.
#define BOARD22(x) ((piece_t *)( (char *)&b->board[2][2] + (x)))
#define VALUE6(p, x) ((float *)( (char *)&b->values[6+p][0] + (x)))
#define ATTACK22(x) ((piece_t *)( (char *)&b->attack[2][2] + (x)))
#ifndef SIMPLEBOARD
//attackvector and len for direction dir at square x, where x is in *void format.
//i'm using the *void pointer format, because it makes address calculations for
//the compiler somewhat easier. unfortunatly is makes the code much harder to read
//because of the ugly typecasts, so i use the macros below.
#define ATTACKVEC22(dir, x) ((unsigned int *)( (char *)&b->attackvec[dir][2][2] + (x)))
#define ATTACKLEN22(dir, x) ((unsigned int *)( (char *)&b->attacklen[dir][2][2] + (x)))
#define IN_CHECK(color) ((color) > 0 ? ATTACK22(POS(KING,0))->b : ATTACK22(POS(-KING,0))->w)
#else
#define IN_CHECK(color) inCheck((constBoard *)b, color)
#endif

#define X(adr)      (((adr)>>2) & 0xf)
#define Y(adr)      ((adr)>>6)
#define ADR(y,x)    ((x)*COL + (y)*ROW)
#define TYPE(y,x)   ((signed)(b->board[(y)+2][(x)+2].all) >> (9+16))
#define VAL(f,y,x)  b->values[6+(f)][y][x]      //value einer Figur auf einem bestimmten feld
#define CNT(x)      b->pieces[6+(x)].cnt        //anzahl einer fiugren art
#define POS(x,y)    b->pieces[6+(x)].pos[y]     //position der y-ten figur von typ x

typedef struct _FullMove {
    Move move;
    int epstatus;
    int o_o;
    int captured;
    int fiftyMoves;
} FullMove;

typedef struct _Board {
//we use the classic 12x10 internal representation, extended to 12x16,
//which fits better into a cache line (64 byte on my machine)
    piece_t board[12][16];
//the value table. first index is biased by 6, so
//value[0-5] = value tables for black king-rook
//value[7-12] = value tables for white rook-king
    float values[13][8][16];
//the attack table. this is the reason we need a 12x16, since squares outside
//the board are accessed to save boundary checks.
    piece_t attack[16][16];
#ifndef SIMPLEBOARD
//in the attack vector table the long range pieces store their attacks, along
//with updating attack[][]. this is neccessary to make blocking attacks work.
    unsigned int attackvec[8][16][16];
//in this table the length of an attack vector from the square to the edge of the board
//or a blocking piece is stored.
    int attacklen[8][16][16];
#endif
//attack signatures
    unsigned pext[13][16][16];    
#define PEXT(p, x) (*((unsigned int *)( (char *)&b->pext[6+p][2][2] + (x))))

//the list of piece positions.
    piecel_t pieces[13];

    unsigned long long repTable[REPTABLESIZE];

    FullMove moveList[MAXGAMELEN], *currMove, *startMove, *lastMove;

    int epstatus;
    int o_o;
    int color;
    int fiftyMoves;

    HashKey key;

    unsigned wsig, bsig;
    int material; //in pawns

} Board;

typedef struct _constBoard {
//we use the classic 12x10 internal representation, extended to 12x16,
//which fits better into a cache line (64 byte on my machine)
    const piece_t board[12][16];
//the value table. first index is biased by 6, so
//value[0-5] = value tables for black king-rook
//value[7-12] = value tables for white rook-king
    const float values[13][8][16];
//the attack table. this is the reason we need a 12x16, since squares outside
//the board are accessed to save boundary checks.
    const piece_t attack[16][16];
#ifndef SIMPLEBOARD
//in the attack vector table the long range pieces store their attacks, along
//with updating attack[][]. this is neccessary to make blocking attacks work.
    const unsigned int attackvec[8][16][16];
//in this table the length of an attack vector from the square to the edge of the board
//or a blocking piece is stored.
    const int attacklen[8][16][16];
#endif
//attack signatures
    const unsigned pext[13][16][16];    
#define PEXT(p, x) (*((unsigned int *)( (char *)&b->pext[6+p][2][2] + (x))))

//the list of piece positions.
    const piecel_t pieces[13];

    const unsigned long long repTable[REPTABLESIZE];

    const FullMove moveList[MAXGAMELEN], *currMove, *startMove, *lastMove;

    int epstatus;
    int o_o;
    int color;
    int fiftyMoves;

    HashKey key;

    unsigned wsig, bsig;
    int material; //in pawns

} constBoard;

static const int dirtab[8] = { COL, ROW + COL, ROW, ROW - COL, -COL, -ROW - COL, -ROW, -ROW + COL };
#define dirtab_(dir) \
(dir==0 ? COL : ( dir==1 ? ROW + COL : ( dir==2 ?  ROW : ( dir==3 ? ROW-COL :\
(dir==4 ?-COL : ( dir==5 ?-ROW - COL : ( dir==6 ? -ROW : -ROW+COL)))))))
static const int kdirtab[8] = { 2 * ROW + COL, ROW + 2 * COL, 2 * ROW - COL, ROW - 2 * COL, -2 * ROW - COL, -ROW - 2 * COL, -2 * ROW + COL, -ROW + 2 * COL };

extern Board board __attribute__ ((aligned(64)));

void convert2EPD(const Board *b, char *);
void integrity();
void initBoard(void);
void benchBoard(Board *);
void clearBoard(Board *);
void doMove(Board *, Move *);
void undoMove(Board *);
int insertRep(Board *);
void deleteRep(Board *);

unsigned long long searchTime(void);
void setTimeout(double, double);
void setStop(int);
void setBreakMove(const Move *);

void sort(Move * movelist, Move *, int color);
void sortw(Move * l, Move * r);
void sortb(Move * l, Move * r);

int putPiece(Board * b, int pos, int piece);
int clrPiece(Board * b, int pos);
int movePiece(Board * b, int src, int dst);
void unmovePiece(Board * b, int src, int dst, int cap);

#endif
