/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>

#include "hash.h"
#include "board.h"
#include "mt19937int.h"
#include "search.h"
#include "debug.h"
#include "util.h"

HashKey hashKeys[13][8][16];

HashInfo mainHash, pawnHash, whiteHash, blackHash, permHash;

char permInfoFileName[256];
char permHashFileName[256];
char permTempFileName[256];

static FILE *permInfoFile;
static FILE *permHashFile;
static FILE *permTempFile;

int permPly = (10*PLY);

//flushes all permanent stuff back to disk
void flushHash() {
    if (permInfoFile) {
	rewind(permInfoFile);
	fwrite(&permHash, sizeof(HashInfo), 1, permInfoFile);
	fflush(permInfoFile);
    }
    if (permTempFile)
	fflush(permTempFile);
}
    
void initHash() {

    mainHash.entrySize = sizeof(MainEntry);
    whiteHash.entrySize = blackHash.entrySize = sizeof(HashEntry);
    pawnHash.entrySize = sizeof(PawnEntry);
    
    permInfoFile = fopen(permInfoFileName, "r+");
    if (permInfoFile)
	fread(&permHash, sizeof(HashInfo), 1, permInfoFile);
    
    permHashFile = fopen(permHashFileName, "r");
    
    permTempFile = fopen(permHashFileName, "a");
    
    memset(hashKeys, 0, sizeof(HashKey));
    sgenrand(1);
    for (int p = -6; p <= 6; p++)
        for (int y = 0; y < 8; y++)
            for (int x = 0; x < 8; x++) 
                if (p) {
                    hashKeys[p+6][y][x].main.key = genrand64();
                    if (abs(p) == PAWN /*|| abs(p) == KING*/)
                        hashKeys[p+6][y][x].pawn.key = genrand64();
                    if (p == ROOK || p == BISHOP || p == QUEEN || p == KNIGHT  || p == -KING)
                        hashKeys[p+6][y][x].white.key = genrand64();
                    if (p == -ROOK || p == -BISHOP || p == -QUEEN || p == -KNIGHT  || p == KING)
                        hashKeys[p+6][y][x].black.key = genrand64();
                }
//add rooks in the startposition and one besides to the pawn hashtable
//needed for castling evaluation
    hashKeys[ROOK+6][0][0].pawn.key = genrand64();                
    hashKeys[ROOK+6][0][1].pawn.key = genrand64();                
    hashKeys[ROOK+6][0][6].pawn.key = genrand64();                
    hashKeys[ROOK+6][0][7].pawn.key = genrand64();                
    hashKeys[-ROOK+6][7][0].pawn.key = genrand64();                
    hashKeys[-ROOK+6][7][1].pawn.key = genrand64();                
    hashKeys[-ROOK+6][7][6].pawn.key = genrand64();                
    hashKeys[-ROOK+6][7][7].pawn.key = genrand64();                
    #if 0
    for (int y = 0; y < 8; y++)
        for (int x = 0; x < 8; x++) {
             hashKeys[PAWN+6][y][x].white.key = genrand64();
             hashKeys[PAWN+6][y][x].black.key = hashKeys[-PAWN+6][y][x].black.key = genrand64();
        }
    #endif
    
}

void allocHash(HashInfo * h, size_t mem) {
    mem = min(mem, (size_t)sysconf(_SC_PAGESIZE) * (size_t)sysconf(_SC_PHYS_PAGES));
retry:
    h->count = 1 << lrint(trunc(log2(mem/h->entrySize)));
    h->size = h->count * h->entrySize;
    
    h->mask = h->count - 1;     //bitmask for fast address calculation of hashelements
    if (h == &mainHash)
        h->mask--;

    free(h->table);
    if ( !(h->table = malloc(h->size)) ) {
        mem >>= 1;
        goto retry;
    }

    bzero(h->table, h->size);
}

void preloadHash() {
    if (!permHashFile)
	return;
    PermEntry temp;
    rewind(permHashFile);
    while (fread(&temp, sizeof(temp), 1, permHashFile)) {
	MainEntry *p = ((MainEntry *) mainHash.table + (temp.key.key0 & mainHash.mask));
	if (temp.ply >= p->ply) {
	    p->key1 = temp.key.key1;
	    p->val = temp.val;
	    p->ply = temp.ply;
	    p->src = temp.src;
	    p->dst = temp.dst;	
	    p->special = temp.special;
	    p->bound = PERM_VALUE;
	    p->overwrite = 1;
	    p->endofsearch = 0;
	}
    }
}

void purgeHash() {
    MainEntry m = {.overwrite = 1 };
    mainHash.fill = 0;

#undef USE_ASM
#ifdef USE_ASM
#ifdef SMALLHASH
    asm("\
        movq    (%2),%%mm0;\
        1:   movntq %%mm0,(%0);\
        movntq %%mm0,0x8(%0);\
        movntq %%mm0,0x10(%0);\
        movntq %%mm0,0x18(%0);\
        sub     $0x4,%1;\
        lea     0x20(%0),%0;\
        jnz     1b;\
    emms;":
    :"r"(mainHash.table), "r"(mainHash.count), "g"(&m)
       );
#else
    asm("\
        movq    (%2),  %%mm0;\
        movq    8(%2), %%mm1;\
   1:   movntq %%mm0,(%0);\
        movntq %%mm1, 0x8(%0);\
        movntq %%mm0,0x10(%0);\
        movntq %%mm1,0x18(%0);\
        movntq %%mm0,0x20(%0);\
        movntq %%mm1,0x28(%0);\
        movntq %%mm0,0x30(%0);\
        movntq %%mm1,0x38(%0);\
        sub     $0x4,%1;\
        lea     0x40(%0),%0;\
        jnz     1b;\
    emms;":
    :"r"(mainHash.table), "r"(mainHash.count), "g"(&m)
       );
#endif       
#else

    for (unsigned int i = 0; i < mainHash.count; i++)
        ((MainEntry *) mainHash.table)[i] = m;
#endif
}

void clearHash() {
#ifdef USE_ASM
    MainEntry m = { .overwrite = 1 };
    int i=-1;
    unsigned *p = (unsigned *)&m;
    while (!p[++i])
        ;
    unsigned a = p[i];

    mainHash.fill = 0;

#ifdef SMALLHASH
    volatile asm("\
        add %3, %0;\
        1:   prefetchw 0x100(%0);\
        or %2, (%0);\
        or %2, 0x8(%0);\
        or %2, 0x10(%0);\
        or %2, 0x18(%0);\
        sub     $0x4,%1;\
        lea     0x20(%0),%0;\
        jnz     1b;"
    :
    :"r"(mainHash.table), "r"(mainHash.count), "r"(a), "r"(i*sizeof(int))
       );
#else
    volatile asm("\
        add %3, %0;\
        1:   prefetchw 0x100(%0);\
        or %2, (%0);\
        or %2, 0x10(%0);\
        or %2, 0x20(%0);\
        or %2, 0x30(%0);\
        sub     $0x4,%1;\
        lea     0x40(%0),%0;\
        jnz     1b;"
    :
    :"r"(mainHash.table), "r"(mainHash.count), "r"(a), "r"(i*sizeof(int))
       );
#endif       
#else

    for (unsigned int i = 0; i < mainHash.count; i++)
        ((MainEntry *) mainHash.table)[i].overwrite = 1;
#endif
}

void writeHash(const struct _constBoard * b, int ply, float val, float alpha, float beta, const Move *m, int endofsearch)
{
    BEGIN_PROFILE(TIME_HASHWRITE);
    MainEntry *p = CURR_HASH_PTR;
    //if neither our ply is sufficent nor the overwrite flag is set, we need to
    //write in the second entry.
//    if (work > 10000000000) asm("int3");
    if (ply < p->ply && !p->overwrite) {
        //if we got a hit in the first entry, don't overwrite the second entry
        //cause this would result in both entrys holding the same key.
        if (p->key1 == b->key.main.key1 ) {
            END_PROFILE();
            return;
        }
        p++;
        stats.write2++;
    } else {
        //if we got neither a overwrite flag nor a hit in the first entry,
        //we can shift the first entry to the second
        if (!p[0].overwrite && p->key1 == b->key.main.key1) {
            if (p[1].overwrite)
                mainHash.fill++;
            p[1] = p[0];
            stats.shift2++;
        }
        mainHash.write++;
    }
    
    if (p->overwrite) {
        mainHash.fill++;
        p->overwrite = 0;
    }
    if (val == -MATE_VALUE) asm("int3");
    p->val = lrint(val * (1.0/QUANT));
//    p->key0 = ((unsigned short*)&b->mainHashKey)[1];
    p->key1 = b->key.main.key1;
    p->ply = ply;
    p->src = m->src >> 2;
    p->endofsearch = endofsearch;
    p->dst = m->dst >> 2;
    p->special = m->special;
    
    if (endofsearch) {
        p->bound = EXACT_VALUE;
    } else {
        if (val <= alpha)
            p->bound = UPPER_BOUND;

        else if (val >= beta)

            p->bound = LOWER_BOUND;

        else
            p->bound = EXACT_VALUE;
    }
    
    if (ply >= permPly && permTempFile) {
	PermEntry temp;
	temp.key.key1 = b->key.main.key1;
	temp.key.key0 = b->key.main.key0 ^ b->color;
	temp.val = p->val;
	temp.ply = p->ply;
	temp.src = p->src;
	temp.dst = p->dst;
	temp.special = p->special;
	fwrite(&temp, sizeof(temp), 1, permTempFile);
    }
    
    END_PROFILE();
}

