/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include "config.h"
#include "main.h"
#include "hash.h"
#include "search.h"
#include "util.h"
#include "opening.h"
#include "movegen_c.h"
#include "debug.h"
#include "parse.h"
#include "smp.h"
#include "squares.h"

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <getopt.h>
#include <errno.h>
#ifdef HAVE_FENV_H
#include <fenv.h>
#endif
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <sys/time.h>

//wrapper for pthread_create, to make gprof work with multithreaded
//applications (like this one).
#ifdef MYDEBUG
    typedef struct wrapper_s {
        void * (*start_routine)(void *);
        void * arg;

        pthread_mutex_t lock;
    pthread_cond_t  wait;

    struct itimerval itimer;

    }
    wrapper_t;

    static void * wrapper_routine(void *)
    ;

    /* Same prototype as pthread_create; use some #define magic to
    * transparently replace it in other files */
    static int gprof_pthread_create(pthread_t * thread, pthread_attr_t * attr,
                            void * (*start_routine)(void *), void * arg) {
        wrapper_t wrapper_data;
        int i_return;

        /* Initialize the wrapper structure */
        wrapper_data.start_routine = start_routine;
        wrapper_data.arg = arg;
        getitimer(ITIMER_PROF, &wrapper_data.itimer);
        pthread_cond_init(&wrapper_data.wait, NULL);
        pthread_mutex_init(&wrapper_data.lock, NULL);
        pthread_mutex_lock(&wrapper_data.lock);

        /* The real pthread_create call */
        i_return = pthread_create(thread, attr, &wrapper_routine,
                                &wrapper_data);

        /* If the thread was successfully spawned, wait for the data
        * to be released */
        if(i_return == 0) {
            pthread_cond_wait(&wrapper_data.wait, &wrapper_data.lock);
        }

        pthread_mutex_unlock(&wrapper_data.lock);
        pthread_mutex_destroy(&wrapper_data.lock);
        pthread_cond_destroy(&wrapper_data.wait);

        return i_return;
    }

    /* The wrapper function in charge for setting the itimer value */
    static void * wrapper_routine(void * data) {
        /* Put user data in thread-local variables */
        void * (*start_routine)(void *) = ((wrapper_t*)data)->start_routine;
        void * arg = ((wrapper_t*)data)->arg;

        /* Set the profile timer value */
        setitimer(ITIMER_PROF, &((wrapper_t*)data)->itimer, NULL);

        /* Tell the calling thread that we don't need its data anymore */
        pthread_mutex_lock(&((wrapper_t*)data)->lock)
        ;
        pthread_cond_signal(&((wrapper_t*)data)->wait);
        pthread_mutex_unlock(&((wrapper_t*)data)->lock)
        ;

        /* Call the real function */
        return start_routine(arg);
    }
#else
    #define gprof_pthread_create(a,b,c,d) pthread_create(a,b,c,d)
#endif

pthread_mutex_t searchingMutex = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t searchingCond = PTHREAD_COND_INITIALIZER;

//time control variables. indices are 0=black, 1=analysis, 2=white.
//time is measured in s. 10^15 s = 30 yars, sufficient for
//a long analysis and anyway in 2^32 s unix systime wraps.
long long timeRemaining[3];     //remaining time for movesRemaining[] moves.
int movesRemaining[3];
//timeRemaining[] and movesRemaining[] get initialized with this
long long initTime[3] = { 300 S, 1000000000000000, 300 S };
int initMoves[3] = { 40, 1, 40 };
double lastTC;                  //when the last time control was
unsigned cpuclock;

int timeRunning = 0;            //indicates who's time is running
bool pondering = false;         //if searchm is pondering atm
bool pondermode = true;         //wether it is allowed to ponder
bool computerTurn = false;

Move mList[MAXMOVES];
Move *eList;
Move returnMove, ponderMove, ponderMoveOld;

char responseBuffer[256];

WINDOW *lineWnd, *statWnd, *boardWnd, *inputWnd, *nodeWnd, *moveWnd;

volatile bool workthreadLocked = false;
volatile bool stopSearch = false;
volatile bool searchFini = false;

enum Mode mode = Normal;
struct Flags flags;

void output(const char *s, ...)
{
    va_list ap;
    va_start(ap, s);
    if (flags.xboard) {
        setbuf(stdout, NULL);
        vprintf(s, ap);
    } else
        vsnprintf(responseBuffer, sizeof(responseBuffer), s, ap);
#ifdef MYDEBUG
    if (flags.debug && DEBUG_IO) 
        debug("out: %s\n", responseBuffer);
#endif
    va_end(ap);
}

void unPonder()
{
    if (pondering) {
        stop();
        board.color = -board.color;
        undoMove(&board);
        pondering = false;
        searchFini = false;
    }
}

void inputMove(const char *s)
{
    Move m;

    if (!str2move(s, &m, mList, eList)) {
        output("illegal move: %s\n", s);
    } else {
        if (!flags.xboard)
            output("move %s accepted\n", s);
        if (pondering) {
            if (ponderMoveOld.special == m.special && ponderMoveOld.src == m.src && ponderMoveOld.dst == m.dst) {
                setTC();
                computerTurn = true;
                pondering = false;
            } else {
                stop();
                searchFini = false;
                deleteRep(&board);
                board.color = -board.color;
                undoMove(&board);
                doMove(&board, &m);
                board.color = -board.color;
                insertRep(&board);
                copyBoard();
                setTC();
                computerTurn = true;
                pondering = false;
                start();
            }
        } else if (stopped() || mode == Analyze || mode == Force || mode == Autoplay || mode == Bench) {
            stop();
            doMove(&board, &m);
            board.color = -board.color;
            insertRep(&board);
            copyBoard();
            setTC();
            computerTurn = true;
            if (mode != Force)
                start();
        }
    }
}

void initTC()
{
    int col;
    for (col = 0; col <= 2; col++) {
        timeRemaining[col] = initTime[col];
        movesRemaining[col] = initMoves[col];
    }
    timeRunning = 0;
}

void setTC()
{

    double tc;
    tc = searchTime();
    if (!--movesRemaining[timeRunning + 1]) {
        movesRemaining[timeRunning + 1] += initMoves[timeRunning + 1];
        timeRemaining[timeRunning + 1] += initTime[timeRunning + 1];
    }
    timeRemaining[timeRunning + 1] -= tc - lastTC;

    if (mode == Analyze || mode == Force)
        timeRunning = 0;
    else
        timeRunning = board.color;

    lastTC = tc;

    t0 = 0.5 * timeRemaining[timeRunning + 1] / (movesRemaining[timeRunning + 1] + 1.0);
    t1 = 2.0 * timeRemaining[timeRunning + 1] / (movesRemaining[timeRunning + 1] + 1.0);
}

bool stopped()
{
    if (pthread_mutex_trylock(&searchingMutex) == EBUSY)
        return false;
    pthread_mutex_unlock(&searchingMutex);
    return true;
}

void stop()
{
    if (!stopped()) {
        stopSearch = true;
        pthread_mutex_lock(&searchingMutex);
        pthread_mutex_unlock(&searchingMutex);
    }
}

void start()
{
    pthread_cond_broadcast(&searchingCond);
}

extern void search();

#include "eval.h"
static void *workthread()
{
  //  profil(profileBuf, sizeof(profileBuf), convert2EPD, 0x8000); 
    pthread_mutex_lock(&searchingMutex);
    workthreadLocked = true;

    initTC();                   //reset clocks
    debug("Wait point 1\n");
    pthread_cond_wait(&searchingCond, &searchingMutex);
    initTC();                   //reset clock after first white move
    setTC();

    debug("Entering work loop\n");
    for (;;) {
        searchm(MAXDEPTH);
        searchFini = true;
        pthread_cond_wait(&searchingCond, &searchingMutex);
    }
    return 0;	//never reached
}

//for use in xboard mode (without ncurses).
//polls stdin for 0.1 sec, and builds a input line.
//calls parse() if it gets a \n. does no output besides parse() 
static void readKeyX() {
    fd_set rfds;
    struct timeval tv;
    int retval;
    static char buf[256];
    static char *pbuf = buf;
    
    /* Watch stdin (fd 0) to see when it has input. */
    FD_ZERO(&rfds);
    FD_SET(0, &rfds);
    tv.tv_sec = 0;
    tv.tv_usec = 100000;
    retval = select(1, &rfds, NULL, NULL, &tv);
    if (retval) {
        int c = 0;
        read(0, &c, 1);
        if (pbuf - buf >= (signed) sizeof(buf) - 3) {
            *pbuf++ = c;
            c = '\n';
        }
        if (c == '\n') {
            *responseBuffer = 0;
            sepParse(buf);
            pbuf = buf;
        } else 
            *pbuf++ = c;
        *pbuf = 0;
    }
}

//for use in interactive mode.
//polls stdin for 0.1 sec and builds a input line,
//calls parse() when it gets a newline.
//updates all windows every 0.1 sec.
static void readKey() {
    static char buf[256];
    static char *pbuf = buf;
    int c;
    if ((c = getch()) != ERR) {
        if (pbuf - buf >= (signed) sizeof(buf) - 3) {
            *pbuf++ = c;
            c = '\n';
        }
        switch (c) {
        case KEY_BACKSPACE:
            if (pbuf > buf)
                *--pbuf = 0;
            break;
        case KEY_UP:
            upNode();
            break;
        case KEY_DOWN:
            downNode();
            break;
        case KEY_LEFT:
            leftNode();
            break;
        case KEY_RIGHT:
            rightNode();
            break;
        case KEY_HOME:
            addPos();
            output("position added");
            break;
        case KEY_END:
            delPos();
            output("position deleted");
            break;
        case '\n':
            *responseBuffer = 0;
            sepParse(buf);
            pbuf = buf;
            if (flags.xboard)
                return;
            break;
        default:
            *pbuf++ = c;
        }
        *pbuf = 0;
    }
    //            printNode(nodeWnd);
    displayBoard(boardWnd);
    attrset(COLOR_PAIR(1));
    printStats(statWnd);
    for (c = -1; c < 2; c += 2) {
        long long tr = timeRemaining[c + 1];
        if (c == timeRunning)
            tr -= searchTime() - lastTC;
        tr /= 1000000;
        mvwprintw(boardWnd, 10 + c, 2, "%c%4d:%02d:%02d", tr > 0 ? ' ' : '-', abs(tr) / 3600, (abs(tr) / 60) % 60, abs(tr) % 60);
    }

    werase(inputWnd);
    wattrset(inputWnd, COLOR_PAIR(2));
    mvwprintw(inputWnd, 0, 0, "%s", responseBuffer);
    wattrset(inputWnd, COLOR_PAIR(1));
    mvwprintw(inputWnd, 1, 0, "slibo: %s", buf);
    printMoveList();
    wrefresh(moveWnd);
    wrefresh(lineWnd);
    wrefresh(boardWnd);
    wrefresh(statWnd);
    wrefresh(inputWnd);
}

//gets called whenever the search is finished.
//in normal mode it actually does output a response move,
//in other modes it may do something different.
static void respond() {
    switch (mode) {
    case Normal:
        if (computerTurn) {
            if (returnMove.src | returnMove.dst) {
                output("move %s\n", moveToStr(&returnMove));
                doMove(&board, &returnMove);
                board.color = -board.color;
                insertRep(&board);
                setTC();
                copyBoard();
                if (pondermode && ponderMove.src | ponderMove.dst) {
                    doMove(&board, &ponderMove);
                    board.color = -board.color;
                    ponderMoveOld = ponderMove;
                    insertRep(&board);
                    pondering = true;
                    t0 = t1 = 0;
                    start();
                }
                computerTurn = false;
            } else {
                output("game has ended.");
            }
        }
    case Force:
    case Analyze:
        break;
    case Autoplay:
        if (returnMove.src | returnMove.dst) {
            output("move %s\n", moveToStr(&returnMove));
            doMove(&board, &returnMove);
            board.color = -board.color;
            insertRep(&board);
            setTC();
            copyBoard();
        } else {
            output("game has ended.");
        }
    case Bench:
        break;
    }
}

//looks if stdin is connected to a pipe.
//in this case we assume we are started with a chess board program
//and activate "xboard mode" and skip all the curses stuff.
static void initNcurses() {
#ifndef HAVE_LIBNCURSES
#warning ncurses library not found.
#endif
    struct stat st;
    fstat(0, &st);
    if (S_ISFIFO(st.st_mode) || S_ISSOCK(st.st_mode) ) {
        debug("Connected to a pipe, not starting interactive mode\n");
        signal(SIGINT, SIG_IGN);
//        signal(SIGTERM, SIG_IGN);
        flags.xboard = true;
        setbuf(stdout, NULL);
        setbuf(stdin, NULL);
    } else {
        debug("ncurses\n");
        initscr();
        start_color();
        init_pair(1, COLOR_WHITE, COLOR_BLACK);
        init_pair(2, COLOR_RED, COLOR_WHITE);
        init_pair(3, COLOR_BLACK, COLOR_RED);
        init_pair(4, COLOR_BLACK, COLOR_YELLOW);
        init_pair(5, COLOR_WHITE, COLOR_RED);
        init_pair(6, COLOR_WHITE, COLOR_YELLOW);
        bkgd(COLOR_PAIR(1) | ' ');

        lineWnd = newwin(LINES - 31, COLS - 20, 29, 0);
        wbkgd(lineWnd, COLOR_PAIR(1) | ' ');
        scrollok(lineWnd, true);

        statWnd = newwin(28, 59, 0, 0);
        wbkgd(statWnd, COLOR_PAIR(1) | ' ');

        inputWnd = newwin(2, COLS, LINES - 2, 0);
        wbkgd(inputWnd, COLOR_PAIR(1) | ' ');

        boardWnd = newwin(16, 16, 0, COLS - 16);
        wbkgd(boardWnd, COLOR_PAIR(1) | ' ');

        moveWnd = newwin(LINES - 18, 20, 18, COLS - 20);
        wbkgd(moveWnd, COLOR_PAIR(1) | ' ');

        nodeWnd = newwin(18, COLS - 60 - 16, 0, 60);
        wbkgd(nodeWnd, COLOR_PAIR(1) | ' ');

        noecho();
        keypad(stdscr, TRUE);
        halfdelay(1);
    }
}

//currently there is only one global which needs to be initialized explicitly.
static void initGlobals() {
    char buf[256];
    cpuclock = 1000.0;
    FILE *f = fopen("/proc/cpuinfo", "r");
    while (fgets(buf, 255, f)) {
        if (sscanf(buf, "cpu MHz         : %d", &cpuclock))
            break;
    }
    fclose(f);
    f = fopen("/etc/sliboenginerc", "r");
    mainHash.size = 0x1000000;
    pawnHash.size = 0x400000;
    whiteHash.size = 0x400000;
    
    strcpy(permInfoFileName, "/usr/local/share/sliboengine/perminfo.bin");
    strcpy(permHashFileName, "/usr/local/share/sliboengine/permhash.bin");
    strcpy(permTempFileName, "/usr/local/share/sliboengine/permtemp.bin");
    
    if (f) 
    while (fgets(buf, 255, f)) {
	sscanf(buf, "permInfoFileName %s", permInfoFileName);
	sscanf(buf, "permHashFileName %s", permHashFileName);
	sscanf(buf, "permTempFileName %s", permTempFileName);
	sscanf(buf, "mainHashSize %d", &mainHash.size);
	sscanf(buf, "pawnHashSize %d", &pawnHash.size);
	sscanf(buf, "pieceHashSize %d", &whiteHash.size);
    }
    blackHash.size = whiteHash.size;
}

int main(int argc, char *argv[]) {
    argc = argc;
#ifdef HAVE_FENV_H    
    feenableexcept(FE_INVALID | FE_DIVBYZERO | FE_OVERFLOW);
#endif  
#ifdef MYDEBUG
    debuglog = fopen("debug.log", "a");
    setlinebuf(debuglog);
#endif
    initGlobals();
    movegenqInit();

    //the main board, the hastables and the workthread need to be initialized
    //before we can start parsing commands (or options)
    initProfiler();
    initEval();
    initHash();
    allocHash(&mainHash, 1<<25);
    allocHash(&pawnHash, 1<<23);
    allocHash(&whiteHash, 1<<23);
    allocHash(&blackHash, 1<<23);
    initBoard();
#ifdef SMP
    initThreads();
#endif
    pthread_t wthread;
    gprof_pthread_create(&wthread, NULL, workthread, NULL);
//    pthread_create(&wthread, NULL, workthread, NULL);

    //we need to wait until workthread has locked its mutex,
    //because stopped() uses this mutex to determine if the
    //workthread has stopped.
    while (!workthreadLocked) ;
    
    //now it is safe to use stopped().
    //wait until workthread has finished initializing and is at a
    //defined postion. 
    while (!stopped()) ;        
    
    initNcurses();
    parse(argv+1);

    flags.log = false;
    debug("Entering main loop\n");
    for (;;) {
        if (flags.xboard) 
            readKeyX();
        else 
            readKey();
        
        if (searchFini) {
            while (!stopped()) ;
            searchFini = false;
            respond();
        }
    }
}
