/***************************************************************************
 *   Copyright (C) 2003 by Gunther Piez                                    *
 *   gpiez@users.sourceforge.net                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 ***************************************************************************/

#include <stdlib.h>
#include "hash.h"
#include "main.h"
#include "debug.h"

static const int matlist[] = {0, 0, 3, 9, 3, 5, 0, 5, 3, 9, 3, 0, 0};
 
#ifdef SIMPLEBOARD

#define putPiece_(dummy, pos, piece) BOARD22(pos)->w = piece << 9; BOARD22(pos)->b = piece << 9;
#define clrPiece_(dummy, pos, piece) BOARD22(pos)->all = 0;
#define chgPiece_(dummy, pos, piece)

#else

#define WROOK_ATTACK ROOK_ATTACK
#define BROOK_ATTACK (ROOK_ATTACK<<16)
#define WQUEEN_ATTACK QUEEN_ATTACK
#define BQUEEN_ATTACK (QUEEN_ATTACK<<16)
#define WBISHOP_ATTACK BISHOP_ATTACK
#define BBISHOP_ATTACK (BISHOP_ATTACK<<16)
#define WKING_ATTACK KING_ATTACK
#define BKING_ATTACK (KING_ATTACK<<16)
#define WPAWN_ATTACK PAWN_ATTACK
#define BPAWN_ATTACK (PAWN_ATTACK<<16)
#define WKNIGHT_ATTACK KNIGHT_ATTACK
#define BKNIGHT_ATTACK (KNIGHT_ATTACK<<16)
//the way it works:
//attacktypes are represented by bit patterns, PAWN..KING_ATTACK. for black this patterns
//are shifted 16 bit to the left. attack[0][sq] stores the sum of all attacks for square sq,
//attackvec[0..7][sq] the patterns for the long range attacking pieces for direction 0..7,
//attacklen[0..7][sq] the length of an range attack, measured from sq to the edge of the board 
//or a piece, including the square of the target piece, not including the source square.
//
// + + + + + + + +
// + q + + + + + +
// + + + + + + + +
// + + + p + R + +
// + + + + + + + +
// + + + + + + + +
// K + + R + + + +
// + + + + + + + b
//
//in this example the poor pawn at d5 is attacked by 2 rooks and defended by a queen and a bishop,
//which yields a mask of (QUEEN_ATTACK+BISHOP_ATTACK) << 16  +  (ROOK_ATTACK+ROOK_ATTACK) at attack[d5]
//attackvec[upwards][d3..d5] contains ROOK_ATTACK, attacklen[upward][d1..d8] is 1,3,2,1,4,3,2,1
//for the length of the attack vectors (=number of free squares til next piece n upwared direction +1).
//when a piece p is placed on position pos, a lookup in attack[0][pos] happens.
//the operation attack[pos] & (ROOK_ATTACK | BISHOP_ATTACK | QUEEN_ATTACK)*7 yields
//the long range attacking pieces, if there are none, the next step is skipped.
//then a lookup for each direction dir (from 1 to 8) in attack[dir][pos] takes place and the difference
//to the new attacktype (ROOK, BISHOP or QUEEN_ATTACK, maybe in a different color) is calculated. if there
//is no difference (because the same type of piece is placed on the same line, happens every time when a range piece
//is moved, because the piece is first placed on the new position and then deleted on the old pos)
//no change to the attack patterns are necessary, and the next step is skipped.
//then the length vectors for the opposite direction are updated in attack[17-dir],
//the new attacktype is written into attack[dir] and the difference is added to attack[0]

//function templates.
//the reason i'm using this ugly function templates for the 8 possible directions is a kind of
//unrolling the loop "for (dir=0; dir<8; dir++)". unfortunatly this loop unrolling is not performed
//by gcc, so i do it on my own. with this optimization expressions like "7*dirtab[dir]"
//and "(dir + 4) & 7" become a constant, so this is pretty useful in terms of performance.

#define ATT(x) ((piece_t *)( (char *)&board.attack[2][2] - (char*)&board + (x)))
#define ATTVEC(dir, x) ((unsigned int *)( (char *)&board.attackvec[dir][2][2] - (char*)&board + (x)))
#define ATTLEN(dir, x) ((unsigned int *)( (char *)&board.attacklen[dir][2][2] - (char*)&board + (x)))

//function template for the eight setAttackVec functions, which are used if a piece is placed
//on the board. setAttackVecN(pos, vec) shortens the lenghth of the attack length in the
//counter direction accordingly and sets attacktype
//vec from square pos in direction N to the next piece or the edge of the board, not including
//the starting square, including the end square. if square was already attacked in that direction,
//the new attackmask is accordingly modified.
#define SETATTACKVEC_FUNC(fname, dir) \
static void fname(unsigned int newAttack, char *pos) __attribute__ ((regparm (2))); \
void fname(unsigned int newAttack, char *pos) { \
    static const void *jmptable[][8] = {{&&set1, &&set2, &&set3, &&set4, &&set5, &&set6, &&set7, &&set8},{&&len1, &&len2, &&len3, &&len4, &&len5, &&len6, &&len7, &&len8}}; \
    int changeLen = *ATTLEN((dir + 4) & 7, pos) + 1; \
    int oldAttack = *ATTVEC(dir, pos); \
    int changeAttack = oldAttack - newAttack; \
    goto *jmptable[changeAttack == 0][*ATTLEN(dir, pos)]; \
set8: \
            ATT(pos + 8*dirtab[dir])->all -= changeAttack; \
            *ATTVEC(dir, pos + 8*dirtab[dir]) = newAttack; \
            *ATTLEN((dir + 4) & 7, pos + 8*dirtab[dir]) -= changeLen; \
set7: \
            ATT(pos + 7*dirtab[dir])->all -= changeAttack; \
            *ATTVEC(dir, pos + 7*dirtab[dir]) = newAttack; \
            *ATTLEN((dir + 4) & 7, pos + 7*dirtab[dir]) -= changeLen; \
set6: \
            ATT(pos + 6*dirtab[dir])->all -= changeAttack; \
            *ATTVEC(dir, pos + 6*dirtab[dir]) = newAttack; \
            *ATTLEN((dir + 4) & 7, pos + 6*dirtab[dir]) -= changeLen; \
set5: \
            ATT(pos + 5*dirtab[dir])->all -= changeAttack; \
            *ATTVEC(dir, pos + 5*dirtab[dir]) = newAttack; \
            *ATTLEN((dir + 4) & 7, pos + 5*dirtab[dir]) -= changeLen; \
set4: \
            ATT(pos + 4*dirtab[dir])->all -= changeAttack; \
            *ATTVEC(dir, pos + 4*dirtab[dir]) = newAttack; \
            *ATTLEN((dir + 4) & 7, pos + 4*dirtab[dir]) -= changeLen; \
set3: \
            ATT(pos + 3*dirtab[dir])->all -= changeAttack; \
            *ATTVEC(dir, pos + 3*dirtab[dir]) = newAttack; \
            *ATTLEN((dir + 4) & 7, pos + 3*dirtab[dir]) -= changeLen; \
set2: \
            ATT(pos + 2*dirtab[dir])->all -= changeAttack; \
            *ATTVEC(dir, pos + 2*dirtab[dir]) = newAttack; \
            *ATTLEN((dir + 4) & 7, pos + 2*dirtab[dir]) -= changeLen; \
set1: \
            ATT(pos + 1*dirtab[dir])->all -= changeAttack; \
            *ATTVEC(dir, pos + 1*dirtab[dir]) = newAttack; \
            *ATTLEN((dir + 4) & 7, pos + 1*dirtab[dir]) -= changeLen; \
    return;\
len8:      *ATTLEN((dir + 4) & 7, pos + 8*dirtab[dir]) -= changeLen; \
len7:      *ATTLEN((dir + 4) & 7, pos + 7*dirtab[dir]) -= changeLen; \
len6:      *ATTLEN((dir + 4) & 7, pos + 6*dirtab[dir]) -= changeLen; \
len5:      *ATTLEN((dir + 4) & 7, pos + 5*dirtab[dir]) -= changeLen; \
len4:      *ATTLEN((dir + 4) & 7, pos + 4*dirtab[dir]) -= changeLen; \
len3:      *ATTLEN((dir + 4) & 7, pos + 3*dirtab[dir]) -= changeLen; \
len2:      *ATTLEN((dir + 4) & 7, pos + 2*dirtab[dir]) -= changeLen; \
len1:      *ATTLEN((dir + 4) & 7, pos + 1*dirtab[dir]) -= changeLen; \
    return;\
}

//function template for the 8 clrAttackVec functions, which are used if a piece
//is removed board.
#define CLRATTACKVEC_FUNC(fname, dir) \
static void fname(unsigned int clrAttack, char *pos) __attribute__ ((regparm (2))); \
void fname(unsigned int clrAttack, char *pos) { \
    static const void *jmptable[][8] = {{&&set1, &&set2, &&set3, &&set4, &&set5, &&set6, &&set7, &&set8},{&&len1, &&len2, &&len3, &&len4, &&len5, &&len6, &&len7, &&len8}}; \
    int oldAttack = *ATTVEC(dir, pos); \
    int changeAttack = oldAttack - clrAttack; \
    int changeLen = *ATTLEN((dir + 4) & 7, pos) + 1; \
    goto *jmptable[changeAttack == 0][*ATTLEN(dir, pos)]; \
set8: \
            ATT(pos + 8*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 8*dirtab[dir]) = oldAttack; \
            *ATTLEN((dir + 4) & 7, pos + 8*dirtab[dir]) += changeLen; \
set7: \
            ATT(pos + 7*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 7*dirtab[dir]) = oldAttack; \
            *ATTLEN((dir + 4) & 7, pos + 7*dirtab[dir]) += changeLen; \
set6: \
            ATT(pos + 6*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 6*dirtab[dir]) = oldAttack; \
            *ATTLEN((dir + 4) & 7, pos + 6*dirtab[dir]) += changeLen; \
set5: \
            ATT(pos + 5*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 5*dirtab[dir]) = oldAttack; \
            *ATTLEN((dir + 4) & 7, pos + 5*dirtab[dir]) += changeLen; \
set4: \
            ATT(pos + 4*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 4*dirtab[dir]) = oldAttack; \
            *ATTLEN((dir + 4) & 7, pos + 4*dirtab[dir]) += changeLen; \
set3: \
            ATT(pos + 3*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 3*dirtab[dir]) = oldAttack; \
            *ATTLEN((dir + 4) & 7, pos + 3*dirtab[dir]) += changeLen; \
set2: \
            ATT(pos + 2*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 2*dirtab[dir]) = oldAttack; \
            *ATTLEN((dir + 4) & 7, pos + 2*dirtab[dir]) += changeLen; \
set1: \
            ATT(pos + 1*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 1*dirtab[dir]) = oldAttack; \
            *ATTLEN((dir + 4) & 7, pos + 1*dirtab[dir]) += changeLen; \
            return;\
len8:      *ATTLEN((dir + 4) & 7, pos + 8*dirtab[dir]) += changeLen; \
len7:      *ATTLEN((dir + 4) & 7, pos + 7*dirtab[dir]) += changeLen; \
len6:      *ATTLEN((dir + 4) & 7, pos + 6*dirtab[dir]) += changeLen; \
len5:      *ATTLEN((dir + 4) & 7, pos + 5*dirtab[dir]) += changeLen; \
len4:      *ATTLEN((dir + 4) & 7, pos + 4*dirtab[dir]) += changeLen; \
len3:      *ATTLEN((dir + 4) & 7, pos + 3*dirtab[dir]) += changeLen; \
len2:      *ATTLEN((dir + 4) & 7, pos + 2*dirtab[dir]) += changeLen; \
len1:      *ATTLEN((dir + 4) & 7, pos + 1*dirtab[dir]) += changeLen; \
    return;\
}

//function template for the 8 chgAttackVec functions, which are used if a piece
//changes on the board (i. e. capture). no change to the attack length happens.
#define CHGATTACKVEC_FUNC(fname, dir) \
static void fname(unsigned int changeAttack, unsigned int newAttack, char *pos) __attribute__ ((regparm (3))); \
void fname(unsigned int changeAttack, unsigned int newAttack, char *pos) { \
    static const void *jmptable[] = {&&set1, &&set2, &&set3, &&set4, &&set5, &&set6, &&set7, &&set8}; \
    goto *jmptable[*ATTLEN(dir, pos)]; \
set8: \
            ATT(pos + 8*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 8*dirtab[dir]) = newAttack; \
set7: \
            ATT(pos + 7*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 7*dirtab[dir]) = newAttack; \
set6: \
            ATT(pos + 6*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 6*dirtab[dir]) = newAttack; \
set5: \
            ATT(pos + 5*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 5*dirtab[dir]) = newAttack; \
set4: \
            ATT(pos + 4*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 4*dirtab[dir]) = newAttack; \
set3: \
            ATT(pos + 3*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 3*dirtab[dir]) = newAttack; \
set2: \
            ATT(pos + 2*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 2*dirtab[dir]) = newAttack; \
set1: \
            ATT(pos + 1*dirtab[dir])->all += changeAttack; \
            *ATTVEC(dir, pos + 1*dirtab[dir]) = newAttack; \
}

static void putPiece_(Board * b, int pos, int piece) __attribute__ ((regparm(3)));
static void clrPiece_(Board * b, int pos, int piece) __attribute__ ((regparm(3)));
static void chgPiece_(Board * b, int pos, int combotype) __attribute__ ((regparm(3)));

//the real function definitions are nested functions, so i don't need to pass *b and pos everywhere.
//putPiece and clrPiece are rather similar, while chgPiece is clearly one of its own kind
//and quite heavy :)

SETATTACKVEC_FUNC(setAttackVec0, 0)
    SETATTACKVEC_FUNC(setAttackVec1, 1)
    SETATTACKVEC_FUNC(setAttackVec2, 2)
    SETATTACKVEC_FUNC(setAttackVec3, 3)
    SETATTACKVEC_FUNC(setAttackVec4, 4)
    SETATTACKVEC_FUNC(setAttackVec5, 5)
    SETATTACKVEC_FUNC(setAttackVec6, 6)
    SETATTACKVEC_FUNC(setAttackVec7, 7)

void putPiece_(Board * b, int pos, int piece)
{

    static const void *jmptable[] = { &&bking, &&bpawn, &&bknight, &&bqueen, &&bbishop, &&brook, &&nothing, &&wrook, &&wbishop, &&wqueen, &&wknight, &&wpawn, &&wking };

    BOARD22(pos)->w = piece << 9;
    BOARD22(pos)->b = piece << 9;
    char *adr = (char *) b + pos;

    goto *jmptable[piece + 6];
  wrook:
    setAttackVec0(WROOK_ATTACK, adr);
    setAttackVec2(WROOK_ATTACK, adr);
    setAttackVec4(WROOK_ATTACK, adr);
    setAttackVec6(WROOK_ATTACK, adr);
    setAttackVec1(0, adr);
    setAttackVec3(0, adr);
    setAttackVec5(0, adr);
    setAttackVec7(0, adr);
    return;
  wbishop:
    setAttackVec0(0, adr);
    setAttackVec2(0, adr);
    setAttackVec4(0, adr);
    setAttackVec6(0, adr);
    setAttackVec1(WBISHOP_ATTACK, adr);
    setAttackVec3(WBISHOP_ATTACK, adr);
    setAttackVec5(WBISHOP_ATTACK, adr);
    setAttackVec7(WBISHOP_ATTACK, adr);
    return;
  wqueen:
    setAttackVec0(WQUEEN_ATTACK, adr);
    setAttackVec2(WQUEEN_ATTACK, adr);
    setAttackVec4(WQUEEN_ATTACK, adr);
    setAttackVec6(WQUEEN_ATTACK, adr);
    setAttackVec1(WQUEEN_ATTACK, adr);
    setAttackVec3(WQUEEN_ATTACK, adr);
    setAttackVec5(WQUEEN_ATTACK, adr);
    setAttackVec7(WQUEEN_ATTACK, adr);
    return;
  wknight:
    ATTACK22(pos + kdirtab[0])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[1])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[2])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[3])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[4])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[5])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[6])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[7])->all += WKNIGHT_ATTACK;
    setAttackVec0(0, adr);
    setAttackVec2(0, adr);
    setAttackVec4(0, adr);
    setAttackVec6(0, adr);
    setAttackVec1(0, adr);
    setAttackVec3(0, adr);
    setAttackVec5(0, adr);
    setAttackVec7(0, adr);
    return;
  wpawn:
    b->key.pawn.key ^= HASHKEY(PAWN, pos).pawn.key;
    ATTACK22(pos + dirtab[1])->all += WPAWN_ATTACK;
    ATTACK22(pos + dirtab[3])->all += WPAWN_ATTACK;
    setAttackVec0(0, adr);
    setAttackVec2(0, adr);
    setAttackVec4(0, adr);
    setAttackVec6(0, adr);
    setAttackVec1(0, adr);
    setAttackVec3(0, adr);
    setAttackVec5(0, adr);
    setAttackVec7(0, adr);
    return;
  wking:
    ATTACK22(pos + dirtab[0])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[1])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[2])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[3])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[4])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[5])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[6])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[7])->all += WKING_ATTACK;
    setAttackVec0(0, adr);
    setAttackVec2(0, adr);
    setAttackVec4(0, adr);
    setAttackVec6(0, adr);
    setAttackVec1(0, adr);
    setAttackVec3(0, adr);
    setAttackVec5(0, adr);
    setAttackVec7(0, adr);
    return;
  nothing:
    debug("putpiece with empty piece\n");
    return;
  brook:
    setAttackVec0(BROOK_ATTACK, adr);
    setAttackVec2(BROOK_ATTACK, adr);
    setAttackVec4(BROOK_ATTACK, adr);
    setAttackVec6(BROOK_ATTACK, adr);
    setAttackVec1(0, adr);
    setAttackVec3(0, adr);
    setAttackVec5(0, adr);
    setAttackVec7(0, adr);
    return;
  bbishop:
    setAttackVec0(0, adr);
    setAttackVec2(0, adr);
    setAttackVec4(0, adr);
    setAttackVec6(0, adr);
    setAttackVec1(BBISHOP_ATTACK, adr);
    setAttackVec3(BBISHOP_ATTACK, adr);
    setAttackVec5(BBISHOP_ATTACK, adr);
    setAttackVec7(BBISHOP_ATTACK, adr);
    return;
  bqueen:
    setAttackVec0(BQUEEN_ATTACK, adr);
    setAttackVec2(BQUEEN_ATTACK, adr);
    setAttackVec4(BQUEEN_ATTACK, adr);
    setAttackVec6(BQUEEN_ATTACK, adr);
    setAttackVec1(BQUEEN_ATTACK, adr);
    setAttackVec3(BQUEEN_ATTACK, adr);
    setAttackVec5(BQUEEN_ATTACK, adr);
    setAttackVec7(BQUEEN_ATTACK, adr);
    return;
  bknight:
    ATTACK22(pos + kdirtab[0])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[1])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[2])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[3])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[4])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[5])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[6])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[7])->all += BKNIGHT_ATTACK;
    setAttackVec0(0, adr);
    setAttackVec2(0, adr);
    setAttackVec4(0, adr);
    setAttackVec6(0, adr);
    setAttackVec1(0, adr);
    setAttackVec3(0, adr);
    setAttackVec5(0, adr);
    setAttackVec7(0, adr);
    return;
  bpawn:
    b->key.pawn.key ^= HASHKEY(-PAWN, pos).pawn.key;
    ATTACK22(pos + dirtab[5])->all += BPAWN_ATTACK;
    ATTACK22(pos + dirtab[7])->all += BPAWN_ATTACK;
    setAttackVec0(0, adr);
    setAttackVec2(0, adr);
    setAttackVec4(0, adr);
    setAttackVec6(0, adr);
    setAttackVec1(0, adr);
    setAttackVec3(0, adr);
    setAttackVec5(0, adr);
    setAttackVec7(0, adr);
    return;
  bking:
    ATTACK22(pos + dirtab[0])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[1])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[2])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[3])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[4])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[5])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[6])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[7])->all += BKING_ATTACK;
    setAttackVec0(0, adr);
    setAttackVec2(0, adr);
    setAttackVec4(0, adr);
    setAttackVec6(0, adr);
    setAttackVec1(0, adr);
    setAttackVec3(0, adr);
    setAttackVec5(0, adr);
    setAttackVec7(0, adr);
    return;
}

CLRATTACKVEC_FUNC(clrAttackVec0, 0)
    CLRATTACKVEC_FUNC(clrAttackVec1, 1)
    CLRATTACKVEC_FUNC(clrAttackVec2, 2)
    CLRATTACKVEC_FUNC(clrAttackVec3, 3)
    CLRATTACKVEC_FUNC(clrAttackVec4, 4)
    CLRATTACKVEC_FUNC(clrAttackVec5, 5)
    CLRATTACKVEC_FUNC(clrAttackVec6, 6)
    CLRATTACKVEC_FUNC(clrAttackVec7, 7)

void clrPiece_(Board * b, int pos, int piece)
{

    static const void *jmptable[] = { &&bking, &&bpawn, &&bknight, &&bqueen, &&bbishop, &&brook, &&nothing, &&wrook, &&wbishop, &&wqueen, &&wknight, &&wpawn, &&wking };

    BOARD22(pos)->all = 0;
    char *adr = (char *) b + pos;

    goto *jmptable[piece + 6];
  wrook:
    clrAttackVec0(WROOK_ATTACK, adr);
    clrAttackVec2(WROOK_ATTACK, adr);
    clrAttackVec4(WROOK_ATTACK, adr);
    clrAttackVec6(WROOK_ATTACK, adr);
    clrAttackVec1(0, adr);
    clrAttackVec3(0, adr);
    clrAttackVec5(0, adr);
    clrAttackVec7(0, adr);
    return;
  wbishop:
    clrAttackVec0(0, adr);
    clrAttackVec2(0, adr);
    clrAttackVec4(0, adr);
    clrAttackVec6(0, adr);
    clrAttackVec1(WBISHOP_ATTACK, adr);
    clrAttackVec3(WBISHOP_ATTACK, adr);
    clrAttackVec5(WBISHOP_ATTACK, adr);
    clrAttackVec7(WBISHOP_ATTACK, adr);
    return;
  wqueen:
    clrAttackVec0(WQUEEN_ATTACK, adr);
    clrAttackVec2(WQUEEN_ATTACK, adr);
    clrAttackVec4(WQUEEN_ATTACK, adr);
    clrAttackVec6(WQUEEN_ATTACK, adr);
    clrAttackVec1(WQUEEN_ATTACK, adr);
    clrAttackVec3(WQUEEN_ATTACK, adr);
    clrAttackVec5(WQUEEN_ATTACK, adr);
    clrAttackVec7(WQUEEN_ATTACK, adr);
    return;
  wknight:
    ATTACK22(pos + kdirtab[0])->all -= WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[1])->all -= WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[2])->all -= WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[3])->all -= WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[4])->all -= WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[5])->all -= WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[6])->all -= WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[7])->all -= WKNIGHT_ATTACK;
    clrAttackVec0(0, adr);
    clrAttackVec2(0, adr);
    clrAttackVec4(0, adr);
    clrAttackVec6(0, adr);
    clrAttackVec1(0, adr);
    clrAttackVec3(0, adr);
    clrAttackVec5(0, adr);
    clrAttackVec7(0, adr);
    return;
  wpawn:
    b->key.pawn.key ^=  HASHKEY(PAWN, pos).pawn.key;
    ATTACK22(pos + dirtab[1])->all -= WPAWN_ATTACK;
    ATTACK22(pos + dirtab[3])->all -= WPAWN_ATTACK;
    clrAttackVec0(0, adr);
    clrAttackVec2(0, adr);
    clrAttackVec4(0, adr);
    clrAttackVec6(0, adr);
    clrAttackVec1(0, adr);
    clrAttackVec3(0, adr);
    clrAttackVec5(0, adr);
    clrAttackVec7(0, adr);
    return;
  wking:
    ATTACK22(pos + dirtab[0])->all -= WKING_ATTACK;
    ATTACK22(pos + dirtab[1])->all -= WKING_ATTACK;
    ATTACK22(pos + dirtab[2])->all -= WKING_ATTACK;
    ATTACK22(pos + dirtab[3])->all -= WKING_ATTACK;
    ATTACK22(pos + dirtab[4])->all -= WKING_ATTACK;
    ATTACK22(pos + dirtab[5])->all -= WKING_ATTACK;
    ATTACK22(pos + dirtab[6])->all -= WKING_ATTACK;
    ATTACK22(pos + dirtab[7])->all -= WKING_ATTACK;
    clrAttackVec0(0, adr);
    clrAttackVec2(0, adr);
    clrAttackVec4(0, adr);
    clrAttackVec6(0, adr);
    clrAttackVec1(0, adr);
    clrAttackVec3(0, adr);
    clrAttackVec5(0, adr);
    clrAttackVec7(0, adr);
    return;
  nothing:
    debug("clrpiece with empty piece\n");
    return;
  brook:
    clrAttackVec0(BROOK_ATTACK, adr);
    clrAttackVec2(BROOK_ATTACK, adr);
    clrAttackVec4(BROOK_ATTACK, adr);
    clrAttackVec6(BROOK_ATTACK, adr);
    clrAttackVec1(0, adr);
    clrAttackVec3(0, adr);
    clrAttackVec5(0, adr);
    clrAttackVec7(0, adr);
    return;
  bbishop:
    clrAttackVec0(0, adr);
    clrAttackVec2(0, adr);
    clrAttackVec4(0, adr);
    clrAttackVec6(0, adr);
    clrAttackVec1(BBISHOP_ATTACK, adr);
    clrAttackVec3(BBISHOP_ATTACK, adr);
    clrAttackVec5(BBISHOP_ATTACK, adr);
    clrAttackVec7(BBISHOP_ATTACK, adr);
    return;
  bqueen:
    clrAttackVec0(BQUEEN_ATTACK, adr);
    clrAttackVec2(BQUEEN_ATTACK, adr);
    clrAttackVec4(BQUEEN_ATTACK, adr);
    clrAttackVec6(BQUEEN_ATTACK, adr);
    clrAttackVec1(BQUEEN_ATTACK, adr);
    clrAttackVec3(BQUEEN_ATTACK, adr);
    clrAttackVec5(BQUEEN_ATTACK, adr);
    clrAttackVec7(BQUEEN_ATTACK, adr);
    return;
  bknight:
    ATTACK22(pos + kdirtab[0])->all -= BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[1])->all -= BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[2])->all -= BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[3])->all -= BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[4])->all -= BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[5])->all -= BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[6])->all -= BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[7])->all -= BKNIGHT_ATTACK;
    clrAttackVec0(0, adr);
    clrAttackVec2(0, adr);
    clrAttackVec4(0, adr);
    clrAttackVec6(0, adr);
    clrAttackVec1(0, adr);
    clrAttackVec3(0, adr);
    clrAttackVec5(0, adr);
    clrAttackVec7(0, adr);
    return;
  bpawn:
    b->key.pawn.key ^= HASHKEY(-PAWN, pos).pawn.key;
    ATTACK22(pos + dirtab[5])->all -= BPAWN_ATTACK;
    ATTACK22(pos + dirtab[7])->all -= BPAWN_ATTACK;
    clrAttackVec0(0, adr);
    clrAttackVec2(0, adr);
    clrAttackVec4(0, adr);
    clrAttackVec6(0, adr);
    clrAttackVec1(0, adr);
    clrAttackVec3(0, adr);
    clrAttackVec5(0, adr);
    clrAttackVec7(0, adr);
    return;
  bking:
    ATTACK22(pos + dirtab[0])->all -= BKING_ATTACK;
    ATTACK22(pos + dirtab[1])->all -= BKING_ATTACK;
    ATTACK22(pos + dirtab[2])->all -= BKING_ATTACK;
    ATTACK22(pos + dirtab[3])->all -= BKING_ATTACK;
    ATTACK22(pos + dirtab[4])->all -= BKING_ATTACK;
    ATTACK22(pos + dirtab[5])->all -= BKING_ATTACK;
    ATTACK22(pos + dirtab[6])->all -= BKING_ATTACK;
    ATTACK22(pos + dirtab[7])->all -= BKING_ATTACK;
    clrAttackVec0(0, adr);
    clrAttackVec2(0, adr);
    clrAttackVec4(0, adr);
    clrAttackVec6(0, adr);
    clrAttackVec1(0, adr);
    clrAttackVec3(0, adr);
    clrAttackVec5(0, adr);
    clrAttackVec7(0, adr);
    return;
}

CHGATTACKVEC_FUNC(chgAttackVec0, 0)
    CHGATTACKVEC_FUNC(chgAttackVec1, 1)
    CHGATTACKVEC_FUNC(chgAttackVec2, 2)
    CHGATTACKVEC_FUNC(chgAttackVec3, 3)
    CHGATTACKVEC_FUNC(chgAttackVec4, 4)
    CHGATTACKVEC_FUNC(chgAttackVec5, 5)
    CHGATTACKVEC_FUNC(chgAttackVec6, 6)
    CHGATTACKVEC_FUNC(chgAttackVec7, 7)
//combotype: 16*(newpiece+6) + (oldpiece+6)
void chgPiece_(Board * b, int pos, int combotype)
{

#define chgLine(a,b) \
    chgAttackVec0(a, b, adr); \
    chgAttackVec2(a, b, adr); \
    chgAttackVec4(a, b, adr); \
    chgAttackVec6(a, b, adr);

#define chgDia(a,b) \
    chgAttackVec1(a, b, adr); \
    chgAttackVec3(a, b, adr); \
    chgAttackVec5(a, b, adr); \
    chgAttackVec7(a, b, adr);

#define cbqueen \
        chgDia(-BQUEEN_ATTACK, 0);\
        chgLine(-BQUEEN_ATTACK, 0);

#define cbrook \
        chgLine(-BROOK_ATTACK, 0);

#define cbbishop\
        chgDia(-BBISHOP_ATTACK, 0);

    static void cbknight() {
        ATTACK22(pos + kdirtab[0])->all -= BKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[1])->all -= BKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[2])->all -= BKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[3])->all -= BKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[4])->all -= BKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[5])->all -= BKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[6])->all -= BKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[7])->all -= BKNIGHT_ATTACK;
    }
    static void cbpawn() {
        ATTACK22(pos + kdirtab[5])->all -= BPAWN_ATTACK;
        ATTACK22(pos + kdirtab[7])->all -= BPAWN_ATTACK;
    }

#define cwqueen \
        chgDia(-WQUEEN_ATTACK, 0);\
        chgLine(-WQUEEN_ATTACK, 0);

#define cwrook \
        chgLine(-WROOK_ATTACK, 0);

#define cwbishop \
        chgDia(-WBISHOP_ATTACK, 0);

    static void cwknight() {
        ATTACK22(pos + kdirtab[0])->all -= WKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[1])->all -= WKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[2])->all -= WKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[3])->all -= WKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[4])->all -= WKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[5])->all -= WKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[6])->all -= WKNIGHT_ATTACK;
        ATTACK22(pos + kdirtab[7])->all -= WKNIGHT_ATTACK;
    }
    static void cwpawn() {
        ATTACK22(pos + kdirtab[1])->all -= WPAWN_ATTACK;
        ATTACK22(pos + kdirtab[3])->all -= WPAWN_ATTACK;
    }

    static const void *jmptable[][16] = {
        {0, 0, 0, 0, 0, 0, 0, &&bking_rook, &&bking_bishop, &&bking_queen, &&bking_knight, &&bking_pawn},
        {0, 0, 0, 0, 0, 0, 0, &&bpawn_rook, &&bpawn_bishop, &&bpawn_queen, &&bpawn_knight, &&bpawn_pawn},
        {0, 0, 0, 0, 0, 0, 0, &&bknight_rook, &&bknight_bishop, &&bknight_queen, &&bknight_knight, &&bknight_pawn},
        {0, 0, 0, 0, 0, 0, 0, &&bqueen_rook, &&bqueen_bishop, &&bqueen_queen, &&bqueen_knight, &&bqueen_pawn},
        {0, 0, 0, 0, 0, 0, 0, &&bbishop_rook, &&bbishop_bishop, &&bbishop_queen, &&bbishop_knight, &&bbishop_pawn},
        {0, 0, 0, 0, 0, 0, 0, &&brook_rook, &&brook_bishop, &&brook_queen, &&brook_knight, &&brook_pawn},
        {0},
        {0, &&wrook_pawn, &&wrook_knight, &&wrook_queen, &&wrook_bishop, &&wrook_rook},
        {0, &&wbishop_pawn, &&wbishop_knight, &&wbishop_queen, &&wbishop_bishop, &&wbishop_rook},
        {0, &&wqueen_pawn, &&wqueen_knight, &&wqueen_queen, &&wqueen_bishop, &&wqueen_rook},
        {0, &&wknight_pawn, &&wknight_knight, &&wknight_queen, &&wknight_bishop, &&wknight_rook},
        {0, &&wpawn_pawn, &&wpawn_knight, &&wpawn_queen, &&wpawn_bishop, &&wpawn_rook},
        {0, &&wking_pawn, &&wking_knight, &&wking_queen, &&wking_bishop, &&wking_rook},
    };
#ifdef MYDEBUG
    int low = combotype % 16;
    int high = combotype / 16;
    if ((low > 6 && high > 6) || (low < 6 && high < 6) || (low > 12 || low < 0 || high > 12 || high < 0)) {
        debug("illegal combotype %d %d\n", high, low);
        return;
    }
#endif
    char *adr = (char *) b + pos;
    goto *jmptable[0][combotype];

  wrook_rook:
    chgLine(WROOK_ATTACK - BROOK_ATTACK, WROOK_ATTACK);
    return;
  wrook_bishop:
    cbbishop;
    goto wrook;
  wrook_queen:
    chgDia(-BQUEEN_ATTACK, 0);
    chgLine(WROOK_ATTACK - BQUEEN_ATTACK, WROOK_ATTACK);
    return;
  wrook_knight:
    cbknight();
    goto wrook;

  wbishop_rook:
    cbrook;
    goto wbishop;
  wbishop_bishop:
    chgDia(WBISHOP_ATTACK - BBISHOP_ATTACK, WBISHOP_ATTACK);
    return;
  wbishop_queen:
    chgLine(-BQUEEN_ATTACK, 0);
    chgDia(WBISHOP_ATTACK - BQUEEN_ATTACK, WBISHOP_ATTACK);
    return;
  wbishop_knight:
    cbknight();
    goto wbishop;

  wqueen_rook:
    chgLine(WQUEEN_ATTACK - BROOK_ATTACK, WQUEEN_ATTACK);
    chgDia(WQUEEN_ATTACK, 0);
    return;
  wqueen_bishop:
    chgDia(WQUEEN_ATTACK - BBISHOP_ATTACK, WQUEEN_ATTACK);
    chgLine(WQUEEN_ATTACK, 0);
    return;
  wqueen_queen:
    chgDia(WQUEEN_ATTACK - BQUEEN_ATTACK, WQUEEN_ATTACK);
    chgLine(WQUEEN_ATTACK - BQUEEN_ATTACK, WQUEEN_ATTACK);
    return;
  wqueen_knight:
    cbknight();
    goto wqueen;

  wknight_rook:
    cbrook;
    goto wknight;
  wknight_bishop:
    cbbishop;
    goto wknight;
  wknight_queen:
    cbqueen;
    goto wknight;
  wknight_knight:
    ATTACK22(pos + kdirtab[0])->all += WKNIGHT_ATTACK - BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[1])->all += WKNIGHT_ATTACK - BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[2])->all += WKNIGHT_ATTACK - BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[3])->all += WKNIGHT_ATTACK - BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[4])->all += WKNIGHT_ATTACK - BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[5])->all += WKNIGHT_ATTACK - BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[6])->all += WKNIGHT_ATTACK - BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[7])->all += WKNIGHT_ATTACK - BKNIGHT_ATTACK;
    return;

  wpawn_rook:
    cbrook;
    goto wpawn;
  wpawn_bishop:
    cbbishop;
    goto wpawn;
  wpawn_queen:
    cbqueen;
    goto wpawn;
  wpawn_knight:
    cbknight();
    goto wpawn;

  wking_rook:
    cbrook;
    goto wking;
  wking_bishop:
    cbbishop;
    goto wking;
  wking_queen:
    cbqueen;
    goto wking;
  wking_knight:
    cbknight();
    goto wking;

  wking_pawn:
    cbpawn();
  wking:
    ATTACK22(pos + dirtab[0])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[1])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[2])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[3])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[4])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[5])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[6])->all += WKING_ATTACK;
    ATTACK22(pos + dirtab[7])->all += WKING_ATTACK;
    return;

  wknight_pawn:
    cbpawn();
  wknight:
    ATTACK22(pos + kdirtab[0])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[1])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[2])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[3])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[4])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[5])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[6])->all += WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[7])->all += WKNIGHT_ATTACK;
    return;

  wpawn_pawn:
    cbpawn();
  wpawn:
    ATTACK22(pos + dirtab[1])->all += WPAWN_ATTACK;
    ATTACK22(pos + dirtab[3])->all += WPAWN_ATTACK;
    return;

  wbishop_pawn:
    cbpawn();
  wbishop:
    chgDia(WBISHOP_ATTACK, WBISHOP_ATTACK);
    return;

  wqueen_pawn:
    cbpawn();
  wqueen:
    chgLine(WQUEEN_ATTACK, WQUEEN_ATTACK);
    chgDia(WQUEEN_ATTACK, WQUEEN_ATTACK);
    return;

  wrook_pawn:
    cbpawn();
  wrook:
    chgLine(WROOK_ATTACK, WROOK_ATTACK);
    return;

  brook_rook:
    chgLine(BROOK_ATTACK - WROOK_ATTACK, BROOK_ATTACK);
    return;
  brook_bishop:
    cwbishop;
    goto brook;
  brook_queen:
    chgDia(-WQUEEN_ATTACK, 0);
    chgLine(BROOK_ATTACK - WQUEEN_ATTACK, BROOK_ATTACK);
    return;
  brook_knight:
    cwknight();
    goto brook;

  bbishop_rook:
    cwrook;
    goto bbishop;
  bbishop_bishop:
    chgDia(BBISHOP_ATTACK - WBISHOP_ATTACK, BBISHOP_ATTACK);
    return;
  bbishop_queen:
    chgLine(-WQUEEN_ATTACK, 0);
    chgDia(BBISHOP_ATTACK - WQUEEN_ATTACK, BBISHOP_ATTACK);
    return;
  bbishop_knight:
    cwknight();
    goto bbishop;

  bqueen_rook:
    chgLine(BQUEEN_ATTACK - WROOK_ATTACK, BQUEEN_ATTACK);
    chgDia(BQUEEN_ATTACK, 0);
    return;
  bqueen_bishop:
    chgDia(BQUEEN_ATTACK - WBISHOP_ATTACK, BQUEEN_ATTACK);
    chgLine(BQUEEN_ATTACK, 0);
    return;
  bqueen_queen:
    chgDia(BQUEEN_ATTACK - WQUEEN_ATTACK, BQUEEN_ATTACK);
    chgLine(BQUEEN_ATTACK - WQUEEN_ATTACK, BQUEEN_ATTACK);
    return;
  bqueen_knight:
    cwknight();
    goto bqueen;

  bknight_rook:
    cwrook;
    goto bknight;
  bknight_bishop:
    cwbishop;
    goto bknight;
  bknight_queen:
    cwqueen;
    goto bknight;
  bknight_knight:
    ATTACK22(pos + kdirtab[0])->all += BKNIGHT_ATTACK - WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[1])->all += BKNIGHT_ATTACK - WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[2])->all += BKNIGHT_ATTACK - WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[3])->all += BKNIGHT_ATTACK - WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[4])->all += BKNIGHT_ATTACK - WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[5])->all += BKNIGHT_ATTACK - WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[6])->all += BKNIGHT_ATTACK - WKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[7])->all += BKNIGHT_ATTACK - WKNIGHT_ATTACK;
    return;

  bpawn_rook:
    cwrook;
    goto bpawn;
  bpawn_bishop:
    cwbishop;
    goto bpawn;
  bpawn_queen:
    cwqueen;
    goto bpawn;
  bpawn_knight:
    cwknight();
    goto bpawn;

  bking_rook:
    cwrook;
    goto bking;
  bking_bishop:
    cwbishop;
    goto bking;
  bking_queen:
    cwqueen;
    goto bking;
  bking_knight:
    cwknight();
    goto bking;

  bking_pawn:
    cwpawn();
  bking:
    ATTACK22(pos + dirtab[0])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[1])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[2])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[3])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[4])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[5])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[6])->all += BKING_ATTACK;
    ATTACK22(pos + dirtab[7])->all += BKING_ATTACK;
    return;

  bknight_pawn:
    cwpawn();
  bknight:
    ATTACK22(pos + kdirtab[0])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[1])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[2])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[3])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[4])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[5])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[6])->all += BKNIGHT_ATTACK;
    ATTACK22(pos + kdirtab[7])->all += BKNIGHT_ATTACK;
    return;

  bpawn_pawn:
    cwpawn();
  bpawn:
    ATTACK22(pos + dirtab[5])->all += BPAWN_ATTACK;
    ATTACK22(pos + dirtab[7])->all += BPAWN_ATTACK;
    return;

  bbishop_pawn:
    cwpawn();
  bbishop:
    chgDia(BBISHOP_ATTACK, BBISHOP_ATTACK);
    return;

  bqueen_pawn:
    cwpawn();
  bqueen:
    chgLine(BQUEEN_ATTACK, BQUEEN_ATTACK);
    chgDia(BQUEEN_ATTACK, BQUEEN_ATTACK);
    return;

  brook_pawn:
    cwpawn();
  brook:
    chgLine(BROOK_ATTACK, BROOK_ATTACK);
    return;
}
#endif                          //!SIMPLEBOARD

static void printdebug(Board * b)
{
    int x, y, d;
    printf("board\n");
    for (y = 0; y < 8; y++) {
        for (x = 0; x < 8; x++)
            printf("%9ld", b->board[y + 2][x + 2].all >> (9 + 16));
        printf("\n");
    }
#ifndef SIMPLEBOARD
    printf("attack\n");
    for (y = 0; y < 8; y++) {
        for (x = 0; x < 8; x++)
            printf("%9x", (unsigned) b->attack[y + 2][x + 2].all);
        printf("\n");
    }
    for (d = 0; d < 8; d++) {
        printf("attackvec %d\n", d);
        for (y = 0; y < 8; y++) {
            for (x = 0; x < 8; x++)
                printf("%9x", b->attackvec[d][y + 2][x + 2]);
            printf("\n");
        }
    }
    for (d = 0; d < 8; d++) {
        printf("attacklen %d\n", d);
        for (y = 0; y < 8; y++) {
            for (x = 0; x < 8; x++)
                printf("%9x", b->attacklen[d][y + 2][x + 2]);
            printf("\n");
        }
    }
#endif
    printf("----------------------------------------------------------------\n");
}

#ifdef USE_ASM
static inline void xor1(Board *b, const HashKey *k1) {
    asm volatile ("\
        movq   (%0), %%mm0;\
        movq  8(%0), %%mm1;\
        movq 16(%0), %%mm2;\
        movq 24(%0), %%mm3;\
        pxor   (%1), %%mm0;\
        pxor  8(%1), %%mm1;\
        pxor 16(%1), %%mm2;\
        pxor 24(%1), %%mm3;\
        movq %%mm0,   (%0);\
        movq %%mm1,  8(%0);\
        movq %%mm2, 16(%0);\
        movq %%mm3, 24(%0);\
        emms"
        :
        :"r" (&b->key), "r" (k1));
}
        
static inline void xor2(Board *b, const HashKey *k1, const HashKey *k2) {
    asm volatile ("\
        movq   (%0), %%mm0;\
        movq  8(%0), %%mm1;\
        movq 16(%0), %%mm2;\
        movq 24(%0), %%mm3;\
        pxor   (%1), %%mm0;\
        pxor  8(%1), %%mm1;\
        pxor 16(%1), %%mm2;\
        pxor 24(%1), %%mm3;\
        pxor   (%2), %%mm0;\
        pxor  8(%2), %%mm1;\
        pxor 16(%2), %%mm2;\
        pxor 24(%2), %%mm3;\
        movq %%mm0,   (%0);\
        movq %%mm1,  8(%0);\
        movq %%mm2, 16(%0);\
        movq %%mm3, 24(%0);\
        emms"
        :
        :"r" (&b->key), "r" (k1), "r" (k2));
}
        
static inline void xor3(Board *b, const HashKey *k1, const HashKey *k2, const HashKey *k3) {
    asm volatile ("\
        movq   (%0), %%mm0;\
        movq  8(%0), %%mm1;\
        movq 16(%0), %%mm2;\
        movq 24(%0), %%mm3;\
        pxor   (%1), %%mm0;\
        pxor  8(%1), %%mm1;\
        pxor 16(%1), %%mm2;\
        pxor 24(%1), %%mm3;\
        pxor   (%2), %%mm0;\
        pxor  8(%2), %%mm1;\
        pxor 16(%2), %%mm2;\
        pxor 24(%2), %%mm3;\
        pxor   (%3), %%mm0;\
        pxor  8(%3), %%mm1;\
        pxor 16(%3), %%mm2;\
        pxor 24(%3), %%mm3;\
        movq %%mm0,   (%0);\
        movq %%mm1,  8(%0);\
        movq %%mm2, 16(%0);\
        movq %%mm3, 24(%0);\
        emms"
        :
        :"r" (&b->key), "r" (k1), "r" (k2), "r" (k3));
}
#endif
        
//highlevel functions.
//highlevel functions.
//they update the piece lists and the hashtables.

int putPiece(Board * b, int pos, int piece)
{
    int cap = BOARD22(pos)->all >> (16 + 9);
#ifdef MYDEBUG
    if (flags.debug & DEBUG_MOVES)
        debug("    putpiece pos %x piece %d cap %d\n", pos, piece, cap);
    if (piece == 0) {
        debug("piece error\n");
        asm("int3");
    }
#endif
    //  printdebug(b);

    if (cap) {
        int *ppp = &POS(cap, --CNT(cap));
        if (*ppp != pos) {
            int temp2 = *ppp;
            while (*--ppp != pos) {
                int temp1 = *ppp;
                *ppp = temp2;
                temp2 = temp1;
            }
            *ppp = temp2;
        }
        POS(piece, CNT(piece)++) = pos;
        #ifdef USE_ASM
        xor2(b, &HASHKEY(piece, pos), &HASHKEY(cap, pos));
        #else
        b->key.main.key ^= HASHKEY(piece, pos).main.key ^ HASHKEY(cap, pos).main.key;
        b->key.pawn.key ^= HASHKEY(piece, pos).pawn.key ^ HASHKEY(cap, pos).pawn.key;
        b->key.white.key ^= HASHKEY(piece, pos).white.key ^ HASHKEY(cap, pos).white.key;
        b->key.black.key ^= HASHKEY(piece, pos).black.key ^ HASHKEY(cap, pos).black.key;
        #endif
        //combotype: 16*(newpiece+6) + (oldpiece+6)
        chgPiece_(b, pos, cap + 6 + 16 * (piece + 6));
        BOARD22(pos)->b = piece << 9;
        BOARD22(pos)->w = piece << 9;
        b->material += matlist[piece+6] - matlist[cap+6];
    } else {
        POS(piece, CNT(piece)++) = pos;
        #ifdef USE_ASM
        xor1(b, &HASHKEY(piece, pos));
        #else
        b->key.main.key ^= HASHKEY(piece, pos).main.key;
        b->key.pawn.key ^= HASHKEY(piece, pos).pawn.key;
        b->key.white.key ^= HASHKEY(piece, pos).white.key;
        b->key.black.key ^= HASHKEY(piece, pos).black.key;
        #endif
        putPiece_(b, pos, piece);
        b->material += matlist[piece+6];
    }
    if (flags.debug) {
        //printdebug(b);
    }
    return cap;
}

int clrPiece(Board * b, int pos)
{
    int piece = BOARD22(pos)->all >> (16 + 9);
#ifdef MYDEBUG
    if (flags.debug & DEBUG_MOVES)
        debug("    clrpiece pos %x piece %d\n", pos, piece);
    if (piece == 0) {
        debug("piece error\n");
        asm("int3");
    }
#endif
    int *ppp = &POS(piece, --CNT(piece));
    if (*ppp != pos) {
        int temp2 = *ppp;
        while (*--ppp != pos) {
            int temp1 = *ppp;
            *ppp = temp2;
            temp2 = temp1;
        }
        *ppp = temp2;
    }
    #ifdef USE_ASM
    xor1(b, &HASHKEY(piece, pos));
    #else
    b->key.main.key ^= HASHKEY(piece, pos).main.key;
    b->key.pawn.key ^= HASHKEY(piece, pos).pawn.key;
    b->key.white.key ^= HASHKEY(piece, pos).white.key;
    b->key.black.key ^= HASHKEY(piece, pos).black.key;
    #endif
    clrPiece_(b, pos, piece);
    b->material -= matlist[piece+6];
    if (flags.debug) {
        //printdebug(b);
    }
    return piece;
}

int movePiece(Board * b, int src, int dst)
{
    int piece = BOARD22(src)->all >> (16 + 9);
    int cap = BOARD22(dst)->all >> (16 + 9);
#ifdef MYDEBUG
    if (flags.debug & DEBUG_MOVES)
        debug("    movepiece %d %2x-%2x (cap %d)\n", piece, src, dst, cap);
    if (abs(cap) == KING) {
        debug("king captured %x-%X\n", src / 4, dst / 4);
        asm("int3");
    }
    if (piece == 0) {
        debug("piece error\n");
        asm("int3");
    }
#endif
    int *ppp = &POS(piece, CNT(piece));
    while (*--ppp != src) ;
    *ppp = dst;

    if (cap) {
        ppp = &POS(cap, --CNT(cap));
        if (*ppp != dst) {
            int temp2 = *ppp;
            while (*--ppp != dst) {
                int temp1 = *ppp;
                *ppp = temp2;
                temp2 = temp1;
            }
            *ppp = temp2;
        }
        #ifdef USE_ASM
        xor3(b, &HASHKEY(piece, src), &HASHKEY(piece, dst), &HASHKEY(cap, dst));
        #else
        b->key.main.key ^= HASHKEY(piece, src).main.key ^ HASHKEY(piece, dst).main.key ^ HASHKEY(cap, dst).main.key;
        b->key.pawn.key ^= HASHKEY(piece, src).pawn.key ^ HASHKEY(piece, dst).pawn.key ^ HASHKEY(cap, dst).pawn.key;
        b->key.white.key ^= HASHKEY(piece, src).white.key ^ HASHKEY(piece, dst).white.key ^ HASHKEY(cap, dst).white.key;
        b->key.black.key ^= HASHKEY(piece, src).black.key ^ HASHKEY(piece, dst).black.key ^ HASHKEY(cap, dst).black.key;
        #endif
        clrPiece_(b, src, piece);
        chgPiece_(b, dst, cap + 6 + 16 * (piece + 6));
        BOARD22(dst)->b = piece << 9;
        BOARD22(dst)->w = piece << 9;
        b->material -= matlist[cap+6];
    } else {
        #ifdef USE_ASM
        xor2(b, &HASHKEY(piece, src), &HASHKEY(piece, dst));
        #else
        b->key.main.key ^= HASHKEY(piece, src).main.key ^ HASHKEY(piece, dst).main.key;
        b->key.pawn.key ^= HASHKEY(piece, src).pawn.key ^ HASHKEY(piece, dst).pawn.key;
        b->key.white.key ^= HASHKEY(piece, src).white.key ^ HASHKEY(piece, dst).white.key;
        b->key.black.key ^= HASHKEY(piece, src).black.key ^ HASHKEY(piece, dst).black.key;
        #endif
        putPiece_(b, dst, piece);
        clrPiece_(b, src, piece);
    }
    if (flags.debug) {
        //printdebug(b);
    }
    return cap;
}

void unmovePiece(Board * b, int src, int dst, int cap)
{
    int piece = BOARD22(src)->all >> (16 + 9);
#ifdef MYDEBUG
    if (flags.debug & DEBUG_MOVES)
        debug("    unmovepiece %d %2x-%2x (cap %d)\n", piece, src, dst, cap);
    if (piece == 0) {
        debug("piece error\n");
        asm("int3");
    }
#endif
    int *ppp = &POS(piece, CNT(piece));
    while (*--ppp != src) ;
    *ppp = dst;
    if (cap) {
        POS(cap, CNT(cap)++) = src;
        #ifdef USE_ASM
        xor3(b, &HASHKEY(piece, src), &HASHKEY(piece, dst), &HASHKEY(cap, src));
        #else
        b->key.main.key ^= HASHKEY(piece, src).main.key ^ HASHKEY(piece, dst).main.key ^ HASHKEY(cap, src).main.key;
        b->key.pawn.key ^= HASHKEY(piece, src).pawn.key ^ HASHKEY(piece, dst).pawn.key ^ HASHKEY(cap, src).pawn.key;
        b->key.white.key ^= HASHKEY(piece, src).white.key ^ HASHKEY(piece, dst).white.key ^ HASHKEY(cap, src).white.key;
        b->key.black.key ^= HASHKEY(piece, src).black.key ^ HASHKEY(piece, dst).black.key ^ HASHKEY(cap, src).black.key;
        #endif
        putPiece_(b, dst, piece);
        chgPiece_(b, src, piece + 6 + 16 * (cap + 6));
        BOARD22(src)->b = cap << 9;
        BOARD22(src)->w = cap << 9;
        b->material += matlist[cap+6];
    } else {
        #ifdef USE_ASM
        xor2(b, &HASHKEY(piece, src), &HASHKEY(piece, dst));
        #else
        b->key.main.key ^= HASHKEY(piece, src).main.key ^ HASHKEY(piece, dst).main.key;
        b->key.pawn.key ^= HASHKEY(piece, src).pawn.key ^ HASHKEY(piece, dst).pawn.key;
        b->key.white.key ^= HASHKEY(piece, src).white.key ^ HASHKEY(piece, dst).white.key;
        b->key.black.key ^= HASHKEY(piece, src).black.key ^ HASHKEY(piece, dst).black.key;
        #endif
        putPiece_(b, dst, piece);
        clrPiece_(b, src, piece);
    }
    if (flags.debug) {
        //printdebug(b);
    }

}
